from __future__ import annotations

import argparse
import inspect
import re
import sys
import textwrap
from argparse import Action, ArgumentParser, HelpFormatter
from collections.abc import Sequence as AbstractSequence
from functools import lru_cache, partial
from importlib import import_module
from itertools import cycle, zip_longest
from shutil import get_terminal_size
from types import ModuleType
from typing import Optional, Sequence, Tuple, Union
from unittest.mock import patch

from ._actions import OptionalTypeAction
from ._meta import get_concrete_collection_type, is_optional_type

__all__ = ("CorgyHelpFormatter",)

# These placeholders are used to replace special characters and words, so they can be
# identified later for colorizing, without clashes with the help text. The code-points
# are non-character points from `U+FDD0` to `U+FDEF`, which are guaranteed to never be
# used for a character.
_PLACEHOLDER_OPTION_STR = "\ufdd0"
_PLACEHOLDER_METAVAR = "\ufdd1"
_PLACEHOLDER_DEFAULT_VAL = "\ufdd2"
_PLACEHOLDER_EXTRAS_BEGIN = "\ufdd3"
_PLACEHOLDER_EXTRAS_END = "\ufdd4"
_PLACEHOLDER_CHOICES_BEGIN = "\ufdd5"
_PLACEHOLDER_CHOICES_END = "\ufdd6"
_PLACEHOLDER_CHOICES_SEP = "\ufdd7"
_PLACEHOLDER_KWD_DEFAULT = "\ufdd8"
_PLACEHOLDER_KWD_OPTIONAL = "\ufdd9"
_PLACEHOLDER_KWD_REQUIRED = "\ufdda"
_PLACEHOLDER_METAVARS_BEGIN = "\ufddb"
_PLACEHOLDER_METAVARS_END = "\ufddc"
_PLACEHOLDER_METAVARS_REPEAT = "\ufddd"

# These markers are used by `argparse` to indicate metavar sequences, e.g.,
# `[int ...]`, `int [int ...]`.
_MARKER_METAVARS_BEGIN = "["
_MARKER_METAVARS_END = "]"
_MARKER_METAVARS_REPEAT = "..."


class ColorHelper:
    """Wrapper around `crayons` library to colorize text.

    Args:
        use_colors: Whether to enable colored output. If `None`, coloring is enabled
            if the `crayons` library is available, and the output is a tty.
        skip_tty_check: Whether to skip checking if the output is a tty. Only used if
            `use_colors` is None.
    """

    __slots__ = ("crayons",)
    crayons: Optional[ModuleType]

    def __init__(self, use_colors: Optional[bool] = None, skip_tty_check: bool = False):
        if use_colors:
            try:
                self.crayons = import_module("crayons")
            except ImportError:
                raise ImportError(
                    "`crayons` library is required to use colors"
                ) from None
        elif use_colors is None and (skip_tty_check or sys.stdout.isatty()):
            try:
                self.crayons = import_module("crayons")
            except ImportError:
                self.crayons = None
        else:
            self.crayons = None

    def colorize(self, text: str, color: str) -> str:
        """Colorize given text.

        Args:
            text: Text to colorize.
            color: Name of a valid `crayons` color. If the name is all caps, the text
                will be made bold. Special string `BOLD` will only make the text bold,
                without coloring.
        """
        if not self.crayons:
            return text

        if color == "BOLD":
            # `crayons` does not support only making text bold, so we have to use
            # `colorama` directly.
            colorama = getattr(self.crayons, "colorama")
            return colorama.Style.BRIGHT + text + colorama.Style.NORMAL

        use_bold = color.isupper()
        if use_bold:
            color = color.lower()
        try:
            f_color = getattr(self.crayons, color)
        except AttributeError:
            raise ValueError(f"invalid color: {color}") from None
        return str(f_color(text, bold=use_bold))


class _CorgyHelpFormatterMeta(type):
    """Metaclass for `CorgyHelpFormatter` which adds a `__setattr__` method.

    The method prevents new attributes from being set, primarily to prevent potential
    user errors caused by using an incorrect name to configure the class.
    """

    __slots__ = ()

    def __setattr__(cls, name, value):
        # Note: `__setattr__` applies to instances of the class, so `cls` here is a
        # class created using this metaclass.
        if name not in cls.__dict__:
            raise AttributeError(
                f"cannot set attribute `{name}` on class `{cls.__name__}`: if you are"
                f"trying to configure an existing attribute, are you sure you're using"
                f"the correct name?"
            )
        super().__setattr__(name, value)


class CorgyHelpFormatter(HelpFormatter, metaclass=_CorgyHelpFormatterMeta):
    """Formatter class for `argparse` with a cleaner layout, and support for colors.

    `Corgy.parse_from_cmdline` uses this formatter by default, unless a different
    `formatter_class` argument is provided. `CorgyHelpFormatter` can also be used
    independently of `Corgy`. Simply pass it as the `formatter_class` argument to
    `argparse.ArgumentParser()`::

        >>> import argparse
        >>> from argparse import ArgumentParser
        >>> from corgy import CorgyHelpFormatter

        >>> parser = ArgumentParser(
        ...     formatter_class=CorgyHelpFormatter,
        ...     usage=argparse.SUPPRESS,
        ... )
        >>> _ = parser.add_argument("--x", type=int, required=True)
        >>> _ = parser.add_argument("--y", type=str, nargs="*", required=True)
        >>> parser.print_help()
        options:
          -h/--help      show this help message and exit
          --x int        (required)
          --y [str ...]  (required)

    To configure `CorgyHelpFormatter`, you can set a number of attributes on the class.
    Note that you do not need to create an instance of the class; that is done by the
    parser itself. The following public attributes are available:

    * `enable_colors`: If `None` (the default), colors are enabled if the `crayons`
      package is available, and the output is a tty. To explicitly enable or disable
      colors, set to `True` or `False`.

    * `color_<choices/keywords/metavars/defaults/options>`: These attributes control
      the colors used for various parts of the output (see below for reference).
      Available colors are `red`, `green`, `yellow`, `blue`, `black`, `magenta`, `cyan`,
      and `white`. Specifying the name in all caps will make the color bold. You can
      also use the special value `BOLD` to make the output bold without changing the
      color. The default value are `blue` for choices, `green` for keywords, `RED` for
      metavars, `YELLOW` for defaults, and `BOLD` for options. Format:

      .. code-block:: text
         :dedent: 1

              -a/--arg str       help for arg ({'a'/'b'/'c'} default: 'a')
                |      |                          |            |      |
              options  metavars                 choices      keywords defaults

    * `output_width`: The number of columns used for the output. If `None` (the
      default), the current terminal width is used.

    * `max_help_position`: How far to the right (from the start), the help string can
      start from. If `None`, there is no limit. The default is to use half the current
      terminal width.

    * `marker_extras_<begin/end>`: The strings used to enclose the extra help text
      (choices, default values etc.). The defaults are `(` and `)`.

    * `marker_choices_<begin/end>`: The strings used to enclose the list of choices for
      an argument. The defaults are `{` and `}`.

    * `marker_choices_sep`: The string used to separate individual choices in the choice
      list. The default is `/`.

    * `show_full_help`: Whether to show the full help, including choices, indicators for
      required arguments, and the usage string. The default is `True`.

    Formatting of individual arguments can be customized with magic attributes defined
    on the argument type. The following attributes are recognized:

    * `__metavar__`: This can be set to a string on the argument type to override the
        default metavar. Example::

            >>> class T:
            ...     __metavar__ = "METAVAR"

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )
            >>> _ = parser.add_argument("--arg", type=T)
            >>> parser.print_help()
            options:
              --arg METAVAR  (default: None)

    """

    use_colors: Optional[bool] = None
    color_choices = "blue"
    color_defaults = "YELLOW"
    color_keywords = "green"
    color_metavars = "RED"
    color_options = "BOLD"

    output_width: Optional[int] = None
    max_help_position: Optional[int] = get_terminal_size().columns // 2

    marker_extras_begin = "("
    marker_extras_end = ")"
    marker_choices_begin = "{"
    marker_choices_end = "}"
    marker_choices_sep = "/"

    show_full_help = True

    __slots__ = ("_color_helper",)
    _color_helper: ColorHelper

    @property
    def using_colors(self) -> bool:
        """Whether colors are enabled."""
        return self._color_helper.crayons is not None

    # Regex to match a choice within a choice list, e.g., `b` in `{a/b/c}`.
    _pattern_choice = re.compile(
        f"(?<={_PLACEHOLDER_CHOICES_BEGIN}|{_PLACEHOLDER_CHOICES_SEP}).*?"
        f"(?={_PLACEHOLDER_CHOICES_END}|{_PLACEHOLDER_CHOICES_SEP})",
        re.DOTALL,
    )

    # Regex to match any character which is not a metavar extra.
    _pattern_not_metavar_extra = re.compile(
        f"[^{_PLACEHOLDER_METAVARS_BEGIN}{_PLACEHOLDER_METAVARS_END}"
        f"{_PLACEHOLDER_METAVARS_REPEAT}]"
    )

    @staticmethod
    @lru_cache(maxsize=None)
    def _pattern_placeholder_text(placeholder: str) -> re.Pattern:
        """Regex to match text which has been replaced by the given placeholder."""
        # Due to wrapping, the placeholder text may be split across multiple lines. So,
        # the regex looks for a continuous string of `placeholder` or whitespace.
        return re.compile(rf"({placeholder}[{placeholder}\s]*)", re.DOTALL)

    @staticmethod
    def _stringify(obj, type_) -> str:
        if isinstance(obj, (AbstractSequence, set)) and not isinstance(
            obj, (str, bytes)
        ):
            # `obj` is a collection: recursively apply `_stringify` on its elements.
            # `obj` has to be checked before `type_`, because the type may not be a
            # collection type when `nargs` is used to get multiple arguments.
            _coll_type = get_concrete_collection_type(type_)
            if _coll_type is not None and isinstance(
                getattr(type_, "__args__", None), AbstractSequence
            ):
                # `type_` is also a collection, so unwrap it to get the base type. This
                # happens in case of nested types like `Sequence[Sequence[int]]`.
                _base_types = type_.__args__
                if len(_base_types) == 2 and _base_types[1] is Ellipsis:
                    _base_types = _base_types[:1]
                elif len(_base_types) > 1 and len(_base_types) != len(obj):
                    raise ValueError(f"cannot format object as type '{type_}': {obj}")
            else:
                _base_types = [type_]
            _part_strs = [
                CorgyHelpFormatter._stringify(_part, _base_type)
                for _part, _base_type in zip_longest(
                    obj, _base_types, fillvalue=_base_types[-1]
                )
            ]
            _seq_start = "(" if _coll_type is tuple or isinstance(obj, tuple) else "["
            _seq_end = ")" if _seq_start == "(" else "]"
            return _seq_start + ", ".join(_part_strs) + _seq_end

        if is_optional_type(type_):
            # type_ is `Optional`; so unwrap to get the base type. This case happens
            # in cases like `Sequence[Optional[int]]`, where `Optional` is not the
            # outermost type.
            if obj is None:
                return "None"
            return CorgyHelpFormatter._stringify(obj, type_.__args__[0])

        try:
            return obj.__name__  # type: ignore
        except AttributeError:
            try:
                if isinstance(obj, type_):
                    return type_.__str__(obj)
            except TypeError:
                pass
            return str(obj)

    @staticmethod
    def _get_stringify_type_for_default(action):
        """Get the type that should be used to stringify `action.default`."""
        _stringify_type = action.type
        if isinstance(action.nargs, int) or action.nargs in (
            argparse.ZERO_OR_MORE,
            argparse.ONE_OR_MORE,
        ):
            # If the argument specifies nargs, and the default value is a,
            # collection, wrap the action type with the default collection type.
            if isinstance(action.default, tuple):
                _stringify_type = Tuple[action.type]  # type: ignore
            elif isinstance(action.default, AbstractSequence):
                _stringify_type = Sequence[action.type]  # type: ignore
        return _stringify_type

    def _sub_non_ws_with_colored_repl(
        self, match: re.Match, replacement: Optional[str], color: str
    ) -> str:
        """Replace non-whitespace characters in the match using colored replacement.

        For example, if the match is `aaa   aaaa a`, and the replacement is `bbbbbbbb`,
        the result will be `bbb   bbbb b` (with the `b`s colored).

        Args:
            match: The match to substitute into.
            replacement: The replacement to use. If it is shorter than the
                non-whitespace part of the match, it is repeated. If it is `None`, the
                match is replaced with a colored version of itself.
        """
        text = match.group(0)
        text_pieces = re.split(r"(\S+)", text)
        repl_idx = 0
        for i, text_piece in enumerate(text_pieces):
            # Since we split on non-whitespace, every other piece is text.
            if i % 2:
                if replacement is None:
                    repl_piece = text_piece
                else:
                    repl_piece = ""
                    while True:
                        rem_len = len(text_piece) - len(repl_piece)
                        if rem_len <= 0:
                            break
                        repl_piece += replacement[repl_idx : repl_idx + rem_len]
                        repl_idx += rem_len
                        if repl_idx >= len(replacement):
                            repl_idx = 0
                text_pieces[i] = self._color_helper.colorize(repl_piece, color)
        return "".join(text_pieces)

    @staticmethod
    def _get_default_metavar_for_type(type_, using_colors) -> str:
        """Metavar to use if none is explicitly provided.

        Special attribute `__metavar__` can be added to any type, to use a custom
        metavar for that type. Other types use the name of type itself.
        """
        if type_:
            custom_metavar = getattr(type_, "__metavar__", None)
            if custom_metavar is not None:
                return custom_metavar

            if using_colors:
                marker_metavars_begin = _PLACEHOLDER_METAVARS_BEGIN
                marker_metavars_end = _PLACEHOLDER_METAVARS_END
                marker_metavars_repeat = _PLACEHOLDER_METAVARS_REPEAT
            else:
                marker_metavars_begin = _MARKER_METAVARS_BEGIN
                marker_metavars_end = _MARKER_METAVARS_END
                marker_metavars_repeat = _MARKER_METAVARS_REPEAT

            _coll_type = get_concrete_collection_type(type_)
            if _coll_type is not None and (
                isinstance(getattr(type_, "__args__", None), AbstractSequence)
                and type_ is not Sequence
            ):
                # `action.type` is a collection. So, create a metavar list based on the
                # base type(s).
                _type_args = getattr(type_, "__args__")
                if len(_type_args) == 1 or (
                    len(_type_args) == 2 and _type_args[1] is Ellipsis
                ):
                    _base_metavar = CorgyHelpFormatter._get_default_metavar_for_type(
                        _type_args[0], using_colors
                    )
                    # '[<base_type> ...]'.
                    _metavar_repeat = (
                        marker_metavars_begin
                        + _base_metavar
                        + " "
                        + marker_metavars_repeat
                        + marker_metavars_end
                    )
                    if len(_type_args) == 1:
                        return _metavar_repeat
                    # '<base_type> [<base_type> ...]'.
                    return _base_metavar + " " + _metavar_repeat

                _part_metavars = []
                for _part_type in _type_args:
                    _part_metavars.append(
                        CorgyHelpFormatter._get_default_metavar_for_type(
                            _part_type, using_colors
                        )
                    )
                # '<base_type_1> <base_type_2> <...> <base_type_n>'.
                return " ".join(_part_metavars)

            if is_optional_type(type_):
                # `action.type` is optional. So, return '[<base metavar>]'.
                _base_type = getattr(type_, "__args__")[0]
                _s = (
                    marker_metavars_begin
                    + CorgyHelpFormatter._get_default_metavar_for_type(
                        _base_type, using_colors
                    )
                    + marker_metavars_end
                )
                return _s

            try:
                return getattr(type_, "__name__")
            except AttributeError:
                return str(type_)
        return ""

    def _get_default_metavar_for_optional(self, action: Action) -> str:
        return self._get_default_metavar_for_type(action.type, self.using_colors)

    def _format_action_invocation(self, action: Action) -> str:
        """Format the invocation part of an argument, e.g. `-x, --x int`."""
        if action.option_strings:
            option_strings = action.option_strings
        else:
            # If no option strings are present, (positional arguments), use
            # `action.dest`. However, this can be `argparse.SUPPRESS` for sub-actions,
            # in which case use the word `CMD`.
            option_strings = [
                action.dest if action.dest != argparse.SUPPRESS else "CMD"
            ]

        if self.using_colors:
            # Create placeholders for the option strings, and store originals.
            placeholder_option_strings: Sequence[str] = [
                _PLACEHOLDER_OPTION_STR * len(option_string)
                for option_string in option_strings
            ]
            setattr(action, "_corgy_option_strings", option_strings)
        else:
            placeholder_option_strings = option_strings

        # Combine the option strings so that they are shown like `-s/--long ARGS`,
        # rather than `-s ARGS, --long ARGS` (the default).
        with patch.object(
            action,
            "option_strings",
            [self.marker_choices_sep.join(placeholder_option_strings)],
        ):
            return super()._format_action_invocation(action).rstrip()

    def _format_args(self, action: Action, default_metavar: Optional[str]) -> str:
        """Format the metavars."""
        if action.nargs == argparse.PARSER:
            # No metavars for a sub-command.
            return ""

        metavar = action.metavar or default_metavar or ""

        if self.using_colors:
            # Create a placeholder for the metavar, and store it in the action.
            placeholder_metavar: Union[str, Tuple[str, ...]]

            def _placeholderize_metavar(_m):
                # Colors should not be applied to metavar extras, e.g., in
                # `[int ...]`, only `int` must be colored. So, this function
                # replaces non-extra characters with `_PLACEHOLDER_METAVAR`, and
                # returns the modified metavar and the replaced characters.
                _modded, _repl = [], []
                for _char in _m:
                    if _char in [
                        _PLACEHOLDER_METAVARS_BEGIN,
                        _PLACEHOLDER_METAVARS_END,
                        _PLACEHOLDER_METAVARS_REPEAT,
                        " ",
                    ]:
                        _modded.append(_char)
                    else:
                        _modded.append(_PLACEHOLDER_METAVAR)
                        _repl.append(_char)
                return "".join(_modded), "".join(_repl)

            if isinstance(metavar, tuple):
                placeholder_metavar, metavar = zip(
                    *tuple(_placeholderize_metavar(metavar_i) for metavar_i in metavar)
                )
            else:
                placeholder_metavar, metavar = _placeholderize_metavar(metavar)
            # Store the non-extra characters in the action, so they can be colorized
            # and substituted into the placeholder later.
            setattr(action, "_corgy_metavar", metavar)
        else:
            placeholder_metavar = metavar

        # For `corgy._corgy.OptionalTypeAction`s, use the true `nargs` for formatting.
        _fmt_nargs = (
            action._base_nargs
            if isinstance(action, OptionalTypeAction)
            else action.nargs
        )
        with patch.multiple(action, nargs=_fmt_nargs, metavar=placeholder_metavar):
            if action.nargs == argparse.ZERO_OR_MORE:
                # Python 3.9+ shows '*' argumets of a single type as `[<base_type> ...]`
                # instead of `[<base_type> [<base_type> ...]]`. This code backports that
                # functionality.
                _mv = self._metavar_formatter(action, "")(1)
                if len(_mv) == 2:
                    return f"[{_mv[0]} [{_mv[1]} ...]]"
                return f"[{_mv[0]} ...]"
            _fmt = super()._format_args(action, "")
            if isinstance(action, OptionalTypeAction):
                # Add `[]` around the metavar.
                _fmt = f"[{_fmt}]"
            return _fmt

    def _format_action(self, action: Action) -> str:
        """Format a single argument.

        The superclass implementation produces an output like
        `  --arg ARG   arg help`.

        This implementation includes the default value, and choices (if present). Text
        is added to indicate if the argument is required, or optional. The argument type
        is used as the metavar, and colors are used for semantic highlighting.
        """
        # First, generate base format without help text. This is the invocation part,
        # e.g., `--x str`, but with the correct amount of spacing appended, for proper
        # alignment with the other arguments. The help is replaced with a dummy `\0`,
        # since with an empty help, there would be no extra spacing added.
        with patch.object(action, "help", "\0"):
            base_fmt = super()._format_action(action)
        base_fmt = base_fmt[:-2]  # remove trailing `\0\n`

        # Create formatted choice list.
        if action.choices and self.show_full_help:
            if self.using_colors:
                marker_choices_begin = _PLACEHOLDER_CHOICES_BEGIN
                marker_choices_end = _PLACEHOLDER_CHOICES_END
                marker_choices_sep = _PLACEHOLDER_CHOICES_SEP
            else:
                marker_choices_begin = self.marker_choices_begin
                marker_choices_end = self.marker_choices_end
                marker_choices_sep = self.marker_choices_sep
            choices_str = marker_choices_sep.join(
                [self._stringify(choice, action.type) for choice in action.choices]
            )
            choice_list_fmt = (
                marker_choices_begin + choices_str + marker_choices_end + " "
            )
        else:
            choice_list_fmt = ""

        # Compute qualifier (`required`/`optional`/`default`).
        if not action.option_strings:
            # The argument is positional, so it can't be optional, and doesn't have a
            # default value. No extra help text is required, and the extra space added
            # above is removed here.
            if choice_list_fmt:
                choice_list_fmt = choice_list_fmt[:-1]
            arg_qualifier = ""
        elif action.required:
            if self.show_full_help:
                arg_qualifier = (
                    _PLACEHOLDER_KWD_REQUIRED * len("required")
                    if self.using_colors
                    else "required"
                )
            else:
                arg_qualifier = ""
        elif action.default is argparse.SUPPRESS:
            if action.nargs == 0:
                # The argument takes no values, so no need to explicitly indicate that
                # it is optional. For example, help and version actions are obviously
                # optional.
                arg_qualifier = ""
            else:
                arg_qualifier = (
                    _PLACEHOLDER_KWD_OPTIONAL * len("optional")
                    if self.using_colors
                    else "optional"
                )
        else:
            _stringify_type = self._get_stringify_type_for_default(action)
            if self.using_colors:
                arg_qualifier = (
                    (_PLACEHOLDER_KWD_DEFAULT * len("default"))
                    + ": "
                    + (
                        _PLACEHOLDER_DEFAULT_VAL
                        * len(self._stringify(action.default, _stringify_type))
                    )
                )
            else:
                arg_qualifier = (
                    f"default: {self._stringify(action.default, _stringify_type)}"
                )

        # Add qualifier to choice list e.g. `({a/b/c} required)`.
        if choice_list_fmt or arg_qualifier:
            if self.using_colors:
                marker_extras_begin = _PLACEHOLDER_EXTRAS_BEGIN
                marker_extras_end = _PLACEHOLDER_EXTRAS_END
            else:
                marker_extras_begin = self.marker_extras_begin
                marker_extras_end = self.marker_extras_end
            extra_help = (
                marker_extras_begin
                + choice_list_fmt
                + arg_qualifier
                + marker_extras_end
            )
        else:
            extra_help = ""

        # Wrap the text according to `output_width` and `max_help_position`.
        output_width = self.output_width or get_terminal_size().columns
        max_help_position = self.max_help_position or output_width
        if len(base_fmt) + self._current_indent < min(output_width, max_help_position):
            # Example of desired result:
            #     --arg str  the help begins here,
            #                and continue here,
            #                and ends here (optional)
            indent = " " * len(base_fmt)
            base_after_wrap = False  # see below
        else:
            # In this case, the help text cannot start on the first line. So, the help
            # is separately wrapped, and the base format is preprended to it. Example:
            #     --very-very-long-arg-name str
            #         the help begins on the next line
            #         and can extend up to `help_width`.
            #         (optional)
            indent = " " * (self._current_indent + 2 * self._indent_increment)
            base_after_wrap = True  # whether to prepend `base_fmt` after wrapping

            # Since the base format is prepended, it needs to be separately wrapped
            base_fmt = base_fmt.lstrip()
            base_fmt = textwrap.fill(
                base_fmt,
                width=output_width,
                initial_indent=" " * self._current_indent,
                subsequent_indent=" " * (self._current_indent + self._indent_increment),
                break_on_hyphens=False,
            )

        # Combine the base format with the help string and the choice list.
        fmt = "" if base_after_wrap else base_fmt
        if action.help:
            fmt += action.help
        if extra_help:
            fmt += (" " if action.help else "") + extra_help

        fmt = textwrap.fill(
            fmt,
            width=output_width,
            initial_indent=indent if base_after_wrap else "",
            subsequent_indent=indent,
            break_on_hyphens=False,
        )
        if base_after_wrap:
            fmt = base_fmt + "\n" + fmt
        fmt = fmt + "\n"

        if not self.using_colors:
            return fmt

        # Colorize the keywords.
        for kwd, placeholder_kwd in zip(
            ["default", "optional", "required"],
            [
                _PLACEHOLDER_KWD_DEFAULT,
                _PLACEHOLDER_KWD_OPTIONAL,
                _PLACEHOLDER_KWD_REQUIRED,
            ],
        ):
            pattern = self._pattern_placeholder_text(placeholder_kwd)
            f_sub = partial(
                self._sub_non_ws_with_colored_repl,
                replacement=kwd,
                color=self.color_keywords,
            )
            fmt = pattern.sub(f_sub, fmt)

        # Colorize the option strings.
        option_strings = getattr(action, "_corgy_option_strings", None)
        if option_strings is not None:
            pattern = self._pattern_placeholder_text(_PLACEHOLDER_OPTION_STR)
            for option_string in option_strings:
                f_sub = partial(
                    self._sub_non_ws_with_colored_repl,
                    replacement=option_string,
                    color=self.color_options,
                )
                fmt = pattern.sub(f_sub, fmt, count=1)

        # Colorize the metavars.
        metavars = getattr(action, "_corgy_metavar", None)
        if metavars is not None:
            if isinstance(metavars, str):
                metavars = (metavars,)
            if isinstance(action.nargs, int) and action.nargs > 0:
                # When `nargs` is a number, the metavar part is formatted as a space
                # separated sequence. So, all the metavars are captured together by the
                # regex (which allows whitespace between placeholders). So, we combine
                # the metavars into a single string, which will be passed to
                # `_sub_non_ws_with_colored_repl` to replace the entire metavar part.
                metavars = ("".join(metavars),)

            pattern = self._pattern_placeholder_text(_PLACEHOLDER_METAVAR)

            metavar_iter = iter(cycle(metavars))
            while True:
                match = pattern.search(fmt)
                if not match:
                    break
                metavar = next(metavar_iter)
                match_sub = self._sub_non_ws_with_colored_repl(
                    match, metavar, self.color_metavars
                )
                fmt = fmt[: match.start()] + match_sub + fmt[match.end() :]

        # Colorize the default value.
        if action.default != argparse.SUPPRESS:
            pattern = self._pattern_placeholder_text(_PLACEHOLDER_DEFAULT_VAL)
            _stringify_type = self._get_stringify_type_for_default(action)
            f_sub = partial(
                self._sub_non_ws_with_colored_repl,
                replacement=self._stringify(action.default, _stringify_type),
                color=self.color_defaults,
            )
            fmt = pattern.sub(f_sub, fmt)

        # Colorize the choices.
        f_sub = partial(
            self._sub_non_ws_with_colored_repl,
            replacement=None,  # the actual choices are in the regex match
            color=self.color_choices,
        )
        fmt = self._pattern_choice.sub(f_sub, fmt)

        # Replace placeholders.
        fmt = fmt.translate(
            {
                ord(_PLACEHOLDER_CHOICES_BEGIN): self.marker_choices_begin,
                ord(_PLACEHOLDER_CHOICES_END): self.marker_choices_end,
                ord(_PLACEHOLDER_CHOICES_SEP): self.marker_choices_sep,
                ord(_PLACEHOLDER_EXTRAS_BEGIN): self.marker_extras_begin,
                ord(_PLACEHOLDER_EXTRAS_END): self.marker_extras_end,
                ord(_PLACEHOLDER_METAVARS_BEGIN): _MARKER_METAVARS_BEGIN,
                ord(_PLACEHOLDER_METAVARS_END): _MARKER_METAVARS_END,
                ord(_PLACEHOLDER_METAVARS_REPEAT): _MARKER_METAVARS_REPEAT,
            }
        )

        return fmt

    def start_section(self, heading: Optional[str] = None):
        #: :meta private:
        # 'optional arguments' was changed to 'options' in Python 3.10.
        heading = "options" if heading == "optional arguments" else heading
        super().start_section(heading)

    def add_usage(self, *args, **kwargs):
        #: :meta private:
        if self.show_full_help:
            super().add_usage(*args, **kwargs)
            return

        # Only add usage if called directly from `ArgumentParser.format_usage`. This
        # prevents usage from being shown inside help output when `show_full_help` is
        # `False`.
        current_frame = inspect.currentframe()
        while current_frame:
            if current_frame.f_code.co_name == "format_usage":
                super().add_usage(*args, **kwargs)
                break
            current_frame = current_frame.f_back

    def _format_usage(self, usage: Optional[str], *args, **kwargs) -> str:
        with patch.object(self._color_helper, "crayons", None):
            # Disable colors for usage string.
            fmt = super()._format_usage(usage, *args, **kwargs)

        output_width = self.output_width or get_terminal_size().columns
        # Wrap usage to output width.
        fmt = textwrap.fill(
            fmt,
            width=output_width,
            subsequent_indent=" " * self._indent_increment,
            break_on_hyphens=False,
        )
        return fmt + "\n"

    def __init__(self, prog: str):
        # noqa: D107
        self._color_helper = ColorHelper(self.use_colors)
        # Wrapping is managed by this class, so pass `sys.maxsize` to the superclass.
        super().__init__(prog, max_help_position=sys.maxsize, width=sys.maxsize)

    class ShortHelpAction(Action):
        """`argparse.Action` that displays the short help, and exits."""

        def __call__(self, parser, namespace, values, option_string=None):
            CorgyHelpFormatter.show_full_help = False
            parser.print_help()
            parser.exit()

    class FullHelpAction(Action):
        """`argparse.Action` that displays the full help, and exits."""

        def __call__(self, parser, namespace, values, option_string=None):
            CorgyHelpFormatter.show_full_help = True
            parser.print_help()
            parser.exit()

    @classmethod
    def add_short_full_helps(
        cls,
        parser: ArgumentParser,
        short_help_flags: Sequence[str] = ("-h", "--help"),
        full_help_flags: Sequence[str] = ("--helpfull",),
        short_help_msg: str = "show help message and exit",
        full_help_msg: str = "show full help message and exit",
    ):
        """Add arguments for displaying the short or full help.

        The parser must be created with `add_help=False` to prevent a clash with the
        added arguments.

        Args:
            parser: `ArgumentParser` instance to add the arguments to.
            short_help_flags: Sequence of argument strings for the short help option.
                Default is `("-h", "--help")`.
            full_help_flags: Sequence of argument strings for the full help option.
                Default is `("--helpfull")`.
            short_help_msg: String to describe the short help option. Default is `"show
                help message and exit"`.
            full_help_msg: String to describe the full help option. Default is `"show
                full help message and exit"`.

        Example::

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )
            >>> CorgyHelpFormatter.add_short_full_helps(parser)
            >>> parser.print_help()
            options:
              -h/--help   show help message and exit
              --helpfull  show full help message and exit

        """
        parser.add_argument(
            *short_help_flags,
            nargs=0,
            action=cls.ShortHelpAction,
            help=short_help_msg,
            default=argparse.SUPPRESS,
        )
        parser.add_argument(
            *full_help_flags,
            nargs=0,
            action=cls.FullHelpAction,
            help=full_help_msg,
            default=argparse.SUPPRESS,
        )



import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

# Functions to create temporary files and directories created in the home
# directory, or at a path specified by an environment variable.


@contextmanager
def temp_file_in_home(testcase: TestCase):
    """Get a temporary file (or directory) in the home directory."""
    try:
        d = TemporaryDirectory(dir=Path.home())
        fpath = os.path.join(d.name, "temp.file")
        with open(fpath, "wb"):
            pass
        yield Path(fpath)
    except OSError as e:
        testcase.skipTest(f"could not create temp file in home: {e}")
    finally:
        d.cleanup()


@contextmanager
def temp_dir_in_home(testcase: TestCase):
    """Get a temporary directory in the home directory."""
    try:
        d = TemporaryDirectory(dir=Path.home())
        yield Path(d.name)
    except OSError as e:
        testcase.skipTest(f"could not create temp dir in home: {e}")
    finally:
        d.cleanup()


@contextmanager
def temp_env_var_file(testcase: TestCase):
    """Get a temporary file and set an environment variable to its name."""
    # Get a random string to use as the environment variable name
    env_var = f"CORGY_TEMP_{os.urandom(8).hex()}"
    with TemporaryDirectory() as d:
        fpath = os.path.join(d, "temp.file")
        with open(fpath, "wb"):
            pass
        os.environ[env_var] = fpath
        try:
            yield env_var, Path(fpath)
        finally:
            del os.environ[env_var]


@contextmanager
def temp_env_var_dir(testcase: TestCase):
    """Get a temporary directory and set an environment variable to its name."""
    # Get a random string to use as the environment variable name
    env_var = f"CORGY_TEMP_{os.urandom(8).hex()}"
    with TemporaryDirectory() as d:
        os.environ[env_var] = d
        try:
            yield env_var, Path(d)
        finally:
            del os.environ[env_var]



import sys
from typing import ClassVar
from unittest import TestCase
from unittest.mock import MagicMock

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from corgy import Corgy, corgychecker


class TestCorgyCustomCheckers(TestCase):
    def test_corgychecker_raises_if_not_passed_name(self):
        with self.assertRaises(TypeError):

            @corgychecker
            def spam():
                ...

    def test_corgy_raises_if_corgychecker_target_invalid(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int

                @corgychecker("y")
                @staticmethod
                def check(s):
                    ...

    def test_corgy_raises_if_corgychecker_target_not_annotated(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int
                y = 1

                @corgychecker("y")
                @staticmethod
                def check(s):
                    ...

    def test_corgy_raises_if_corgychecker_target_classvar(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: ClassVar[int]

                @corgychecker("x")
                @staticmethod
                def check(s):
                    ...

    def test_corgychecker_allows_decorating_implicit_staticmethod(self):
        class C(Corgy):
            x: int

            @corgychecker("x")
            def check(s):  # type: ignore # pylint: disable=no-self-argument
                ...

        c = C()
        c.x = 1

    def test_corgychecker_raises_if_decorating_non_staticmethod(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int

                @corgychecker("x")
                @classmethod
                def check(cls, s):
                    ...

    def test_corgychecker_functions_are_callable(self):
        class C(Corgy):
            x: int
            y: int

            @corgychecker("x")
            @staticmethod
            def check_x(s):
                return 10

            @corgychecker("y")
            def check_y(s):  # type: ignore # pylint: disable=no-self-argument
                return 20

        self.assertEqual(C.check_x(1), 10)
        self.assertEqual(C.check_y(2), 20)

    def test_corgychecker_accepts_multiple_arguments(self):
        class C(Corgy):
            x: int
            y: int

            @corgychecker("x", "y")
            @staticmethod
            def check(s):
                ...

        self.assertIs(getattr(C, "__checkers")["x"], getattr(C, "__checkers")["y"])

    def test_corgychecker_decorators_can_be_chained(self):
        class C(Corgy):
            x: int
            y: int

            @corgychecker("x")
            @corgychecker("y")
            @staticmethod
            def check(s):
                ...

        self.assertIs(getattr(C, "__checkers")["x"], getattr(C, "__checkers")["y"])

    def test_corgychecker_is_called_on_setattr(self):
        mock_check = MagicMock()

        class C(Corgy):
            x: int
            y: int
            check = corgychecker("x")(mock_check)

        c = C()
        c.x = 1
        c.y = 2
        mock_check.assert_called_once_with(1)

    def test_corgychecker_is_called_after_setattr(self):
        mock_check = MagicMock()

        class C(Corgy):
            x: int
            check = corgychecker("x")(mock_check)

        c = C()
        with self.assertRaises(ValueError):
            c.x = "1"
        mock_check.assert_not_called()

    def test_setattr_raises_if_corgychecker_raises(self):
        class C(Corgy):
            x: int

            @corgychecker("x")
            @staticmethod
            def check(s):
                if s % 2:
                    raise ValueError

        c = C()
        c.x = 2
        with self.assertRaises(ValueError):
            c.x = 3

    def test_corgy_cls_inherits_custom_checker(self):
        class C(Corgy):
            x: int

            @corgychecker("x")
            @staticmethod
            def check(s):
                if s % 2:
                    raise ValueError

        class D(C):
            ...

        d = D()
        d.x = 2
        with self.assertRaises(ValueError):
            d.x = 1

    def test_corgychecker_works_with_self_type(self):
        class C(Corgy):
            x: int
            c: Self

            @corgychecker("c")
            @staticmethod
            def check(v: Self):
                if v.x % 2:
                    raise ValueError

        c = C()
        c.x = 1
        c.c = C(x=2)
        with self.assertRaises(ValueError):
            c.c = C(x=3)



import sys
from doctest import DocFileSuite

import corgy
import corgy.types

DOCTEST_MODULES = {
    corgy: ["_corgy.py", "_corgychecker.py", "_corgyparser.py", "_helpfmt.py"],
    corgy.types: [
        "_initargs.py",
        "_inputfile.py",
        "_keyvaluepairs.py",
        "_outputfile.py",
        "_subclass.py",
    ],
}

DOCTEST_FILES = ["../README.md"]


def load_tests(loader, tests, ignore):
    for mod, modfiles in DOCTEST_MODULES.items():
        for file in modfiles:
            tests.addTest(DocFileSuite(file, package=mod))

    if sys.version_info < (3, 11):
        # Skip README doctest for Python 3.11+ since `typing_extensions` is
        # not installed as a dependency.
        for file in DOCTEST_FILES:
            tests.addTest(DocFileSuite(file))

    return tests



import pickle
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from corgy.types import KeyValuePairs

from ._pklclasses import PklTestB, PklTestC


class TestKeyValuePairs(TestCase):
    def test_key_value_pairs_init_returns_unique_class(self):
        type_ = KeyValuePairs[str, int]
        self.assertIsNot(type_, KeyValuePairs)

    def test_key_value_pairs_init_caches_calls(self):
        type_ = KeyValuePairs[str, int]
        self.assertIs(type_, KeyValuePairs[str, int])
        self.assertIsNot(type_, KeyValuePairs[int, str])

    def test_key_value_pairs_init_handles_type_being_unhashable(self):
        class M(type):
            __hash__ = None

        class A(metaclass=M):
            ...

        type_ = KeyValuePairs[A, str]
        self.assertIsNot(type_, KeyValuePairs[A, str])
        type_ = KeyValuePairs[str, A]
        self.assertIsNot(type_, KeyValuePairs[str, A])

    def test_key_value_pairs_raises_if_re_subscripted(self):
        type_ = KeyValuePairs[str, int]
        with self.assertRaises(TypeError):
            type_[str, int]  # type: ignore # pylint: disable=pointless-statement

    def test_key_value_pairs_returns_dict_like(self):
        type_ = KeyValuePairs[str, str]
        dic = type_("foo=1,bar=2")
        self.assertDictEqual(dic, {"foo": "1", "bar": "2"})

    def test_key_value_pairs_allows_call_without_init(self):
        dic = KeyValuePairs("foo=1,bar=2")
        self.assertDictEqual(dic, {"foo": "1", "bar": "2"})

    def test_key_value_pairs_handles_empty_string(self):
        self.assertDictEqual(KeyValuePairs(""), {})

    def test_key_value_pairs_raises_if_any_item_not_correct_format(self):
        with self.assertRaises(ValueError):
            KeyValuePairs("foo=1,bar2")

    def test_key_value_pairs_handles_multiple_separators_in_item(self):
        dic = KeyValuePairs("foo==1,bar=2=3")
        self.assertDictEqual(dic, {"foo": "=1", "bar": "2=3"})

    def test_key_value_pairs_handles_type_casting(self):
        type_ = KeyValuePairs[int, float]
        dic = type_("1=2.0,3=4.0")
        self.assertDictEqual(dic, {1: float("2.0"), 3: float("4.0")})

    def test_key_value_pairs_raises_if_type_casting_fails(self):
        type_ = KeyValuePairs[int, int]
        with self.assertRaises(ValueError):
            type_("foo=1")
        with self.assertRaises(ValueError):
            type_("1=foo")

    def test_key_value_pairs_handles_type_casting_with_custom_type(self):
        class A:
            def __init__(self, x):
                self.x = x

            def __eq__(self, other):
                return isinstance(other, A) and self.x == other.x

        type_ = KeyValuePairs[str, A]
        dic = type_("foo=1,bar=2")
        self.assertDictEqual(dic, {"foo": A("1"), "bar": A("2")})

    def test_key_value_pairs_metavar(self):
        self.assertEqual(KeyValuePairs.__metavar__, "key=val,...")
        self.assertEqual(KeyValuePairs[str, int].__metavar__, "key=val,...")

    def test_key_value_pairs_handles_custom_sequence_separator(self):
        type_ = KeyValuePairs[str, str]
        with patch.object(type_, "sequence_separator", ";"):
            dic = type_("foo=1;bar=2")
            self.assertDictEqual(dic, {"foo": "1", "bar": "2"})
            self.assertEqual(type_.__metavar__, "key=val;...")

    def test_key_value_pairs_handles_custom_item_separator(self):
        type_ = KeyValuePairs[str, str]
        with patch.object(type_, "item_separator", ":"):
            dic = type_("foo:1,bar:2")
            self.assertDictEqual(dic, {"foo": "1", "bar": "2"})
            self.assertEqual(type_.__metavar__, "key:val,...")

    def test_key_value_pairs_subtype_not_affected_by_changes_to_base_type(self):
        type_ = KeyValuePairs[str, int]
        with patch.multiple(KeyValuePairs, sequence_separator=";", item_separator=":"):
            self.assertEqual(type_.sequence_separator, ",")
            self.assertEqual(type_.item_separator, "=")

    def test_key_value_pairs_repr_str(self):
        type_ = KeyValuePairs[str, int]
        dic = type_("foo=1,bar=2")
        self.assertEqual(repr(dic), "KeyValuePairs[str,int]('foo=1,bar=2')")
        self.assertEqual(str(dic), "{'foo': 1, 'bar': 2}")

    def test_key_value_pairs_repr_str_with_custom_separators(self):
        with patch.multiple(KeyValuePairs, sequence_separator=";", item_separator=":"):
            dic = KeyValuePairs("foo:1;bar:2")
            self.assertEqual(repr(dic), "KeyValuePairs('foo:1;bar:2')")
            self.assertEqual(str(dic), "{'foo': '1', 'bar': '2'}")

    def test_key_value_pairs_accepts_dict(self):
        dic = KeyValuePairs[str, int]({"foo": 1, "bar": 2})
        self.assertDictEqual(dic, {"foo": 1, "bar": 2})
        self.assertEqual(repr(dic), "KeyValuePairs[str,int]({'foo': 1, 'bar': 2})")

    def test_key_value_pairs_pickleable(self):
        with TemporaryDirectory() as _tmpdir:
            _pkl_path = Path(_tmpdir) / "tmp.pkl"

            _KVT = KeyValuePairs[PklTestB, PklTestC]
            _b1, _b2 = PklTestB(1), PklTestB(2)
            _c1, _c2 = PklTestC(1), PklTestC(2)
            _kv = _KVT({_b1: _c1, _b2: _c2})

            with _pkl_path.open("wb") as _f:
                pickle.dump(_kv, _f)

            with _pkl_path.open("rb") as _f:
                _pkv = pickle.load(_f)

            self.assertDictEqual(_kv, _pkv)



from __future__ import annotations

import sys
from typing import TypeVar

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

__all__ = ("Required", "NotRequired")

# `Required` and `NotRequired` are implemented as `Annotated` types.
# These are used to mark attributes as required or not required.
_T = TypeVar("_T")

REQUIRED = object()
Required = Annotated[_T, REQUIRED]

NOT_REQUIRED = object()
NotRequired = Annotated[_T, NOT_REQUIRED]



from __future__ import annotations

import os
import sys
from functools import partial
from typing import Type, TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


def _expand_mixin_base(class_: Type, mod_new: bool) -> Type:
    """Mixin to add support for expanding user directory and environment variables.

    Two concrete implementations of this mixin are provided: `_expand_with_new` and
    `_expand_with_init`. The former is used for classes that have a custom `__new__`
    method, and the latter is used for classes that have a custom `__init__` method.

    The mixins are used as decorators. For example::

        >>> @_expand_with_new  # doctest: +SKIP
        ... class ReadableFile(Path):
        ...     def __new__(cls, path):
        ...         ...

    """
    # Extra docstring to add to the class.
    doc_extra = """

    User directory and environment variable expansion is performed on the path.
    To disable this behavior, set class attributes `do_expanduser` and `do_expandvars`
    to `False` respectively."""

    class ExpandMixin(class_):
        do_expanduser: bool = True
        do_expandvars: bool = True

        @classmethod
        def _do_expand(cls, path: StrPath) -> StrPath:
            if cls.do_expanduser:
                path = os.path.expanduser(path)
            if cls.do_expandvars:
                path = os.path.expandvars(path)
            return path

        if mod_new:

            def __new__(cls: Type[Self], path: StrPath) -> Self:
                path = cls._do_expand(path)
                return super().__new__(cls, path)

        else:

            def __init__(self: Self, path: StrPath) -> None:
                path = self._do_expand(path)
                super().__init__(path)

    base_doc = getattr(class_, "__doc__", "")
    ExpandMixin.__doc__ = base_doc + doc_extra if base_doc else doc_extra.lstrip()

    # Make `ExpandMixin` look like `class`.
    ExpandMixin.__name__ = class_.__name__
    ExpandMixin.__qualname__ = class_.__qualname__
    ExpandMixin.__module__ = class_.__module__
    ExpandMixin.__slots__ = class_.__slots__

    return ExpandMixin


expand_with_new = partial(_expand_mixin_base, mod_new=True)
expand_with_init = partial(_expand_mixin_base, mod_new=False)



from __future__ import annotations

import sys
from typing import Dict, Generic, Iterator, List, Tuple, Type, TypeVar

__all__ = ("SubClass",)
_T = TypeVar("_T")


class _SubClassMeta(type):
    # Python < 3.9 does not support `classmethod` combined with `property`,
    # so we need to define class properties as properties on the metaclass.
    @property
    def __choices__(cls):
        return cls._choices()


class SubClass(Generic[_T], metaclass=_SubClassMeta):
    """Type representing a sub-class of a given class.

    Example::

        >>> from corgy.types import SubClass

        >>> class Base: ...
        >>> class Sub1(Base): ...
        >>> class Sub2(Base): ...

        >>> BaseSubType = SubClass[Base]   # type for a sub-class of `Base`
        >>> BaseSub = BaseSubType("Sub1")  # sub-class of `Base` named `Sub1`
        >>> base_sub = BaseSub()           # instance of a sub-class of `Base`
        >>> base_sub  # doctest: +SKIP
        <Sub1 object at 0x100ea40a0>

    This class cannot be called directly. It first needs to be associated with a base
    class, using the `SubClass[Base]` syntax. This returns a new `SubClass` type, which
    is associated with `Base`. The returned type is callable, and accepts the name of a
    sub-class of `Base`. So, `SubClass[Base]("Sub1")` returns a `SubClass` type instance
    corresponding to the sub-class `Sub1` of `Base`. Finally, the `SubClass` instance
    can be called to create an instance of the sub-class, e.g.,
    `SubClass[Base]("Sub1")()`.

    This class is useful for creating objects of a generic class, where the concrete
    class is determined at runtime, e.g, by a command-line argument::

        >>> from argparse import ArgumentParser

        >>> parser = ArgumentParser()
        >>> _ = parser.add_argument("--base-subcls", type=SubClass[Base])

        >>> args = parser.parse_args(["--base-subcls", "Sub1"])
        >>> base_obj = args.base_subcls()  # an instance of a sub-class of `Base`

    For further convenience when parsing command-line arguments, the class provides a
    `__choices__` property, which returns a tuple of all valid sub-classes, and can be
    passed as the `choices` argument to `ArgumentParser.add_argument`. Refer to the
    docstring of `__choices__` for more information.

    Args:
        name: Name of the sub-class.

    The behavior of sub-class type identification can be customized by setting class
    attributes (preferably on the type returned by the `[...]` syntax).

    * `allow_base`: If `True`, the base class itself will be allowed as a valid
        sub-class. The default is `False`. Example::

            >>> class Base: ...
            >>> class Sub1(Base): ...
            >>> class Sub2(Base): ...
            >>> T = SubClass[Base]
            >>> T.__choices__
            (SubClass[Base]('Sub1'), SubClass[Base]('Sub2'))
            >>> T.allow_base = True
            >>> T.__choices__
            (SubClass[Base]('Base'), SubClass[Base]('Sub1'), SubClass[Base]('Sub2'))

    * `use_full_names`: If `True`, the name passed to the constructor needs to be the
        full name of a sub-class, given by `cls.__module__ + "." + cls.__qualname__`. If
        `False` (the default), the name needs to just be `cls.__name__`. This is useful
        if the sub-classes are not uniquely identified by just their names.

    * `allow_indirect_subs`: If `True` (the default), indirect sub-classes, i.e.,
        sub-classes of the base through another sub-class, are allowed. If `False`,
        only direct sub-classes of the base are allowed. Example::

            >>> class Base: ...
            >>> class Sub1(Base): ...
            >>> class Sub2(Sub1): ...
            >>> T = SubClass[Base]
            >>> T.__choices__
            (SubClass[Base]('Sub1'), SubClass[Base]('Sub2'))
            >>> T.allow_indirect_subs = False
            >>> T.__choices__
            (SubClass[Base]('Sub1'),)

    Note that the types returned by the `SubClass[...]` syntax are cached using the
    base class type. So all instances of `SubClass[Base]` will return the same type,
    and any attributes set on the type will be shared between all instances.
    """

    # The object cache is initialized inside `__class_getitem__`, so every concrete
    # sub-type has its own object cache. The cache uses the sub-class name, along with
    # the class config attributes as key, so that the cache is invalidated when any
    # of the config attributes change.
    _object_cache: Dict[Tuple[str, bool, bool, bool], "SubClass[_T]"]

    allow_base: bool
    use_full_names: bool
    allow_indirect_subs: bool

    _default_allow_base = False
    _default_use_full_names = False
    _default_allow_indirect_subs = True

    _type_cache: Dict[Type[_T], Type[SubClass[_T]]] = {}

    _base: Type[_T]

    _subcls: Type[_T]
    __metavar__ = "cls"
    __slots__ = ("_subcls",)

    def __class_getitem__(cls, item: Type[_T]) -> Type[SubClass[_T]]:
        if hasattr(cls, "_base"):
            raise TypeError(
                f"cannot further sub-script "
                f"`{cls.__name__}[{cls._subclass_name(cls._base)}]`"
            )
        if not hasattr(item, "__subclasses__"):
            raise TypeError(f"`{item}` is not a valid class")

        try:
            ret_type = cls._type_cache[item]
        except (KeyError, TypeError) as e:
            ret_type = type(
                f"{cls.__name__}[{item.__name__}]",
                (cls,),
                {
                    "allow_base": cls._default_allow_base,
                    "use_full_names": cls._default_use_full_names,
                    "allow_indirect_subs": cls._default_allow_indirect_subs,
                    "_base": item,
                    "_object_cache": {},
                    "__slots__": cls.__slots__,
                },
            )
            if not isinstance(e, TypeError):
                # `TypeError` is raised if `item` is not hashable.
                cls._type_cache[item] = ret_type
            sys.modules[ret_type.__module__].__dict__[ret_type.__name__] = ret_type
        return ret_type

    def __getnewargs__(self):
        return (self._subclass_name(self._subcls),)

    @classmethod
    def _ensure_base_set(cls):
        if not hasattr(cls, "_base"):
            raise TypeError(
                f"`{cls.__name__}` must be associated with a base class first: "
                f"use `{cls.__name__}[<class>]`"
            )

    @classmethod
    def _generate_base_subclasses(cls) -> Iterator[Type[_T]]:
        cls._ensure_base_set()

        def _iter_descendants(c):
            for _s in c.__subclasses__():
                yield _s
                yield from _iter_descendants(_s)

        if cls.allow_base:
            yield cls._base

        if cls.allow_indirect_subs:
            yield from _iter_descendants(cls._base)
        else:
            for base_subcls in cls._base.__subclasses__():
                yield base_subcls

    @classmethod
    def _subclass_name(cls, subcls: Type[_T]) -> str:
        if cls.use_full_names:
            return subcls.__module__ + "." + subcls.__qualname__
        return subcls.__name__

    @property
    def which(self) -> Type[_T]:
        """Return the class represented by the `SubClass` instance."""
        return self._subcls

    @classmethod  # type: ignore
    @property
    def __choices__(cls) -> Tuple[SubClass[_T], ...]:
        """Return a tuple of `SubClass` instances for valid sub-classes of the base.

        Each item in the tuple is an instance of `SubClass`, and corresponds to a valid
        sub-class of the base-class associated with this type.
        """
        # For Sphinx.

    @classmethod
    def _choices(cls) -> Tuple[SubClass[_T], ...]:
        cls._ensure_base_set()
        choices: List["SubClass[_T]"] = []
        for subcls in cls._generate_base_subclasses():
            obj = super().__new__(cls)
            obj._subcls = subcls
            choices.append(obj)
        return tuple(choices)

    def __new__(cls, name: str) -> SubClass[_T]:  # pylint: disable=arguments-differ
        cls._ensure_base_set()

        cache_key = (name, cls.allow_base, cls.allow_indirect_subs, cls.use_full_names)
        try:
            return cls._object_cache[cache_key]
        except KeyError:
            pass

        subcls = None
        for subcls in cls._generate_base_subclasses():
            if cls._subclass_name(subcls) == name:
                break
        else:
            if subcls is None:
                raise ValueError(f"`{cls._base}` has no valid sub-classes")
            raise ValueError(f"invalid sub-class name: `{name}`")

        obj = super().__new__(cls)
        obj._subcls = subcls
        cls._object_cache[cache_key] = obj
        return obj

    def __call__(self, *args, **kwargs) -> _T:
        """Return an instance of the sub-class associated with this type.

        Example::

            >>> class Base: ...
            >>> class Sub1(Base):
            ...     def __init__(self, x):
            ...         print(f"initializing `Sub1` with 'x={x}'")

            >>> BaseSubType = SubClass[Base]
            >>> BaseSub = BaseSubType("Sub1")  # an instance of the `SubClass` type

            >>> base_sub = BaseSub(1)
            initializing `Sub1` with 'x=1'

        """
        return self._subcls(*args, **kwargs)  # type: ignore

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SubClass):
            return False
        return self._subcls is other._subcls

    def __hash__(self) -> int:
        return hash(self._subcls)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._subclass_name(self._subcls)!r})"

    def __str__(self) -> str:
        return self._subclass_name(self._subcls)



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Type, Union
from unittest import TestCase

from corgy.types import InputBinFile, InputTextFile, OutputBinFile, OutputTextFile

from ._specialtmps import temp_env_var_file, temp_file_in_home


class TestFileWrapper:
    # The main class is defined inside to prevent `unittest` from discovering it.
    class TestFile(TestCase):
        type: Union[
            Type[OutputTextFile],
            Type[OutputBinFile],
            Type[InputTextFile],
            Type[InputBinFile],
        ]

        def setUp(self):
            self.tmp_dir = TemporaryDirectory()  # pylint: disable=consider-using-with

        def tearDown(self):
            self.tmp_dir.cleanup()

        def test_file_is_correct_type(self):
            with TemporaryDirectory() as d:
                fpath = os.path.join(d, "foo.file")
                with open(fpath, "wb"):
                    pass
                p = self.type(fpath)
                if issubclass(self.type, (OutputTextFile, OutputBinFile)):
                    p.init()
                self.assertIsInstance(p, self.type)
                p.close()

        def test_file_expands_user(self):
            with temp_file_in_home(self) as fpath:
                f = self.type(os.path.join("~", fpath.relative_to(Path.home())))
                if issubclass(self.type, (OutputTextFile, OutputBinFile)):
                    f.init()
                self.assertEqual(str(f), str(fpath))
                f.close()

        def test_file_does_not_expand_user_if_disabled(self):
            class F(self.type):
                do_expanduser = False

            with temp_file_in_home(self) as fpath:
                if issubclass(self.type, (InputTextFile, InputBinFile)):
                    with self.assertRaises(ValueError):
                        F(os.path.join("~", fpath.name))
                    return
                o = F(os.path.join("~", fpath.name))
                o.init()
                self.assertNotEqual(str(o), str(fpath))
                self.assertEqual(str(o), os.path.join("~", fpath.name))
                o.close()
                # Remove the file created by the test, and the '~' directory.
                os.remove(os.path.join(".", "~", fpath.name))
                os.rmdir(os.path.join(".", "~"))

        def test_file_expands_env_var(self):
            with temp_env_var_file(self) as (env_var, fpath):
                o = self.type(f"${env_var}")
                if issubclass(self.type, (OutputTextFile, OutputBinFile)):
                    o.init()
                self.assertEqual(str(o), str(fpath))
                o.close()

        def test_file_does_not_expand_env_var_if_disabled(self):
            class F(self.type):
                do_expandvars = False

            with temp_env_var_file(self) as (env_var, fpath):
                if issubclass(self.type, (InputTextFile, InputBinFile)):
                    with self.assertRaises(ValueError):
                        F(f"${env_var}")
                    return
                o = F(f"${env_var}")
                o.init()
                self.assertNotEqual(str(o), str(fpath))
                self.assertEqual(str(o), f"${env_var}")
                o.close()
                os.remove(os.path.join(".", f"${env_var}"))



# Global classes for pickle tests.
class PklTestA:
    def __init__(self, x: int):
        self.x = x

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.x == self.x

    def __hash__(self):
        return hash(self.x)


class PklTestB(PklTestA):
    ...


class PklTestC(PklTestA):
    ...



"""Corgy package for elegant data classes."""

from ._annotations import *
from ._corgy import *
from ._corgychecker import *
from ._corgyparser import *
from ._helpfmt import *
from ._version import __version__

# pylint: disable=undefined-variable
__all__ = (
    _annotations.__all__  # type: ignore
    + _corgy.__all__  # type: ignore
    + _corgychecker.__all__  # type: ignore
    + _corgyparser.__all__  # type: ignore
    + _helpfmt.__all__  # type: ignore
)



import pickle
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import skipIf, TestCase

from corgy import Corgy
from corgy.types import InitArgs

from ._pklclasses import PklTestA


class TestInitArgs(TestCase):
    def test_init_args_generates_correct_corgy_class(self):
        class A:
            def __init__(self, x: int, y: str):
                ...

        type_ = InitArgs[A]
        self.assertTrue(issubclass(type_, Corgy))
        self.assertTrue(hasattr(type_, "x"))
        self.assertIsInstance(type_.x, property)
        self.assertTrue(hasattr(type_, "y"))
        self.assertIsInstance(type_.y, property)

    def test_init_args_instance_can_be_used_to_init_class(self):
        class A:
            def __init__(self, x: int, y: str):
                self.x = x
                self.y = y

        type_ = InitArgs[A]
        a_args = type_(x=1, y="2")
        a = A(**a_args.as_dict())
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, "2")

    def test_init_args_raises_if_re_subscripted(self):
        class A:
            def __init__(self, x: int, y: str):
                ...

        type_ = InitArgs[A]
        with self.assertRaises(TypeError):
            type_ = type_[A]  # type: ignore

    def test_init_args_raises_if_missing_annotation(self):
        class A:
            def __init__(self, x: int, y):
                ...

        with self.assertRaises(TypeError):
            _ = InitArgs[A]

    def test_init_args_handles_default_values(self):
        class A:
            def __init__(self, x: int, y: str = "foo"):
                ...

        type_ = InitArgs[A]
        a_args = type_(x=1)
        self.assertDictEqual(a_args.as_dict(), {"x": 1, "y": "foo"})

    @skipIf(sys.version_info < (3, 8), "positional-only parameters require Python 3.8+")
    def test_init_args_raises_if_pos_only_arg_present(self):
        # We need to use `exec` to prevent syntax error in Python 3.7.
        # pylint: disable=exec-used
        exec(
            "class A:\n"
            "    def __init__(self, x: int, /, y: str): ...\n"
            "with self.assertRaises(TypeError):\n"
            "    InitArgs[A]",
            globals(),
            locals(),
        )

    def test_init_args_pickleable(self):
        with TemporaryDirectory() as _tmpdir:
            _pkl_path = Path(_tmpdir) / "tmp.pkl"

            _IAT = InitArgs[PklTestA]
            _iat = _IAT(x=1)

            with _pkl_path.open("wb") as _f:
                pickle.dump(_iat, _f)

            with _pkl_path.open("rb") as _f:
                _iat = pickle.load(_f)

            self.assertEqual(_iat, _IAT(x=1))



from __future__ import annotations

import atexit
import os
import sys
from io import BufferedWriter, FileIO, TextIOWrapper
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath

if sys.version_info >= (3, 9):
    from typing import Literal
else:
    from typing_extensions import Literal
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ._expand import expand_with_init

__all__ = ("OutputTextFile", "OutputBinFile", "LazyOutputTextFile", "LazyOutputBinFile")


def _get_output_stream(name: StrPath, mode: Literal["w", "wb"]) -> FileIO:
    """Open a file for writing (creating folders if necessary)."""
    filedir = os.path.dirname(name)
    if filedir and not os.path.exists(filedir):
        try:
            os.makedirs(filedir)
        except OSError as e:
            raise ValueError(
                f"could not create parent directory for `{name}`: {e}"
            ) from None
    try:
        return FileIO(str(name), mode)
    except OSError as e:
        raise ValueError(f"could not open `{name}`: {e}") from None


@expand_with_init
class OutputTextFile(TextIOWrapper):
    """`TextIOWrapper` sub-class representing an output file.

    Args:
        path: Path to a file.

    The file will be created if it does not exist (including any parent directories),
    and opened in text mode (`w`). Existing files will be truncated. `ValueError`
    is raised if any of the operations fail. An `atexit` handler will be registered to
    close the file on program termination.
    """

    __metavar__ = "file"
    __slots__ = ()

    def __init__(self, path: StrPath):
        stream = _get_output_stream(path, "w")
        buffer = BufferedWriter(stream)
        super().__init__(buffer)
        atexit.register(self.__class__.close, self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.buffer.name!r})"

    def __str__(self) -> str:
        return str(self.buffer.name)

    def init(self):
        """No-op for compatibility with `LazyOutputTextFile`."""

    @classmethod
    def _stdoe_wrapper(cls, f: TextIOWrapper) -> Self:
        obj = cls.__new__(cls)
        TextIOWrapper.__init__(obj, f.buffer, line_buffering=True)
        atexit.register(cls.close, obj)
        return obj

    @classmethod
    def stdout_wrapper(cls) -> Self:
        """`sys.__stdout__` as `OutputTextFile`."""
        return cls._stdoe_wrapper(sys.__stdout__)

    @classmethod
    def stderr_wrapper(cls) -> Self:
        """`sys.__stderr__` as `OutputTextFile`."""
        return cls._stdoe_wrapper(sys.__stderr__)


@expand_with_init
class LazyOutputTextFile(OutputTextFile):
    """`OutputTextFile` sub-class that does not auto-initialize.

    Useful for "default" files, which only need to be created if an alternative is not
    provided. `init` must be called on instances before they can be used.
    """

    __slots__ = ("_path",)

    def __init__(self, path: StrPath):
        # pylint: disable=super-init-not-called
        self._path = path

    def init(self):
        """Initialize the file."""
        super().__init__(self._path)


@expand_with_init
class OutputBinFile(BufferedWriter):
    """Type for an output binary file.

    Args:
        path: Path to a file.

    This class is a thin wrapper around `BufferedWriter` that accepts a path, instead
    of a file stream. The file will be created if it does not exist (including any
    parent directories), and opened in binary mode. Existing files will be truncated.
    `ValueError` is raised if any of the operations fail. An `atexit` handler
    will be registered to close the file on program termination.
    """

    __metavar__ = "file"
    __slots__ = ()

    def __init__(self, path: StrPath):
        stream = _get_output_stream(path, "wb")
        super().__init__(stream)
        atexit.register(self.__class__.close, self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name!r})"

    def __str__(self) -> str:
        return str(self.name)

    def init(self):
        """No-op for compatibility with `LazyOutputBinFile`."""

    @classmethod
    def _stdoe_wrapper(cls, f: TextIOWrapper) -> Self:
        obj = cls.__new__(cls)
        stream = FileIO(f.fileno(), mode="wb")
        BufferedWriter.__init__(obj, stream)
        atexit.register(cls.close, obj)
        return obj

    @classmethod
    def stdout_wrapper(cls) -> Self:
        """`sys.__stdout__` as `OutputBinFile`."""
        return cls._stdoe_wrapper(sys.__stdout__)

    @classmethod
    def stderr_wrapper(cls) -> Self:
        """`sys.__stderr__` as `OutputBinFile`."""
        return cls._stdoe_wrapper(sys.__stderr__)


@expand_with_init
class LazyOutputBinFile(OutputBinFile):
    """`OutputBinFile` sub-class that does not auto-initialize.

    Useful for "default" files, which only need to be created if an alternative is not
    provided. `init` must be called on instances before they can be used.
    """

    __slots__ = ("_path",)

    def __init__(self, path: StrPath):
        # pylint: disable=super-init-not-called
        self._path = path

    def init(self):
        """Initialize the file."""
        super().__init__(self._path)



from functools import partial
from typing import Callable, List, NamedTuple, Union

__all__ = ("corgychecker",)


class CorgyChecker(NamedTuple):
    """Class to represent custom checkers.

    This class is returned by the `@corgychecker` decorator, and is used by `Corgy` to
    keep track of checkers.
    """

    var_names: List[str]
    fcheck: Callable

    def __call__(self, val):
        return self.fcheck(val)


def corgychecker(
    *var_names: str,
) -> Callable[[Union[Callable, CorgyChecker]], CorgyChecker]:
    """Decorate a function as a custom checker for one or more attributes.

    To use a custom function for checking the value of a `Corgy` attribute, use this
    decorator. Checking functions must be static, and should only accept a single
    argument, the value to be checked. They should raise `ValueError` to indicate value
    mismatch. Decorating the function with `@staticmethod` is optional, but prevents
    type errors. `@corgychecker` must be the final decorator in the decorator chain.

    Custom checkers are called _after_ type checking, so the values passed to them will
    be of type corresponding to one of the assigned attributes.

    Args:
        var_names: The attributes associated with the decorated checker.

    Example::

        >>> from corgy import Corgy, corgychecker

        >>> class A(Corgy):
        ...     x: int
        ...     @corgychecker("x")
        ...     @staticmethod
        ...     def check_x(val):
        ...         if val % 2:
        ...             raise ValueError(f"'{val}' is not even")

        >>> a = A()
        >>> a.x = 2
        >>> a.x = 3
        Traceback (most recent call last):
           ...
        ValueError: error setting `x`: '3' is not even

    Multiple attributes can use the same checker, either by chaining `corgychecker`, or
    by passing all attribute names directly::

        >>> from typing import Sequence

        >>> class A(Corgy):
        ...     x: int
        ...     y: float
        ...     z: str
        ...     w: Sequence[int]
        ...     @corgychecker("x")
        ...     @corgychecker("y")
        ...     def check_num(val):
        ...         if val < 0:
        ...             raise ValueError("should be non-negative")
        ...     @corgychecker("z", "w")
        ...     def check_seq(val):
        ...         if len(val) > 10:
        ...             raise ValueError("too long")

    """
    if not all(isinstance(_var_name, str) for _var_name in var_names):
        raise TypeError(
            "corgychecker should be passed the name of arguments: decorate using"
            "@corgychecker(<argument(s)>)"
        )

    def wrapper(var_names, fcheck):
        if isinstance(fcheck, CorgyChecker):
            corgy_checker = fcheck
            corgy_checker.var_names.extend(var_names)
        else:
            if isinstance(fcheck, staticmethod):
                fcheck = fcheck.__func__
            if not callable(fcheck):
                raise TypeError("corgychecker can only decorate static functions")
            corgy_checker = CorgyChecker(list(var_names), fcheck)
        return corgy_checker

    return partial(wrapper, var_names)



from __future__ import annotations

import argparse
import sys
from argparse import (
    _ActionsContainer,
    _StoreConstAction,
    _StoreFalseAction,
    _StoreTrueAction,
    ArgumentParser,
)
from collections import defaultdict
from collections.abc import Sequence as AbstractSequence
from dataclasses import dataclass
from functools import partial
from importlib import import_module
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

if sys.version_info >= (3, 9):
    from typing import Literal
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ._actions import BooleanOptionalAction, OptionalTypeAction
from ._corgyparser import CorgyParserAction
from ._helpfmt import CorgyHelpFormatter
from ._meta import (
    check_val_type,
    CorgyMeta,
    get_concrete_collection_type,
    is_literal_type,
    is_optional_type,
)

__all__ = ("Corgy",)
_T = TypeVar("_T", bound="Corgy")


class Corgy(metaclass=CorgyMeta):
    """Base class for collections of attributes.

    To use, subclass `Corgy`, and declare attributes using type annotations::

        >>> from corgy import Corgy

        >>> class A(Corgy):
        ...     x: int
        ...     y: float

    At runtime, class `A` will have `x`, and `y` as properties, so that the class can be
    used similar to Python dataclasses::

        >>> a = A()
        >>> a.x = 1
        >>> a.x
        1

        >>> a.y
        Traceback (most recent call last):
           ...
        AttributeError: no value available for attribute `y`

        >>> a.y = a.x + 1.1
        >>> a.y
        2.1

        >>> del a.x  # unset x
        >>> a.x
        Traceback (most recent call last):
           ...
        AttributeError: no value available for attribute `x`

    Note that the class's `__init__` method only accepts keyword arguments, and ignores
    arguments without a corresponding attribute. The following are all valid::

        >>> A(x=1, y=2.1)
        A(x=1, y=2.1)

        >>> A(x=1, z=3)  # y is not set, and z is ignored
        A(x=1)

        >>> A(**{"x": 1, "y": 2.1, "z": 3})
        A(x=1, y=2.1)

    Attribute values are type-checked, and `ValueError` is raised on type mismatch::

        >>> a = A(x="1")
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type '<class 'int'>': '1'

        >>> a = A()
        >>> a.x = "1"
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type '<class 'int'>': '1'

        >>> class A(Corgy):
        ...     x: int = "1"
        Traceback (most recent call last):
            ...
        ValueError: default value type mismatch for 'x'

    Any type which supports type checking with `isinstance` can be used as an
    attribute type (along with some special type annotations that are discussed below).
    This includes other corgy classes::

        >>> class A(Corgy):
        ...     x: int
        ...     y: float

        >>> class B(Corgy):
        ...     x: int
        ...     a: A

        >>> b = B(x=1)
        >>> b.a = A()
        >>> b.a.x = 10
        >>> b
        B(x=1, a=A(x=10))

    `Corgy` classes have their `__slots__` set to the annotated attributes. So, if you
    want to use additional attributes not tracked by `Corgy`, define them (and only
    them) in `__slots__`::

        >>> class A(Corgy):
        ...     __slots__ = ("x",)
        ...     y: int

        >>> a = A()
        >>> a.y = 1  # `Corgy` attribute
        >>> a.x = 2  # custom attribute
        >>> a
        A(y=1)

    To allow arbitrary instance attributes, add `__dict__` to `__slots__`. Names added
    through custom `__slots__` are not processed by `Corgy`. Alternatively, to disable
    setting `__slots__` completely, set `corgy_make_slots` to `False` in the class
    definition::

        >>> class A(Corgy, corgy_make_slots=False):
        ...     y: int

        >>> a = A()
        >>> a.y = 1  # `Corgy` attribute
        >>> a.x = 2  # custom attribute
        >>> a
        A(y=1)

    Names marked with the `ClassVar` type will be added as class variables, and will
    not be available as `Corgy` attributes::

        >>> from typing import ClassVar

        >>> class A(Corgy):
        ...     x: ClassVar[int] = 3

        >>> A.x
        3
        >>> A.x = 4
        >>> A.x
        4
        >>> a = A()
        >>> a.x
        4
        >>> a.x = 5
        Traceback (most recent call last):
            ...
        AttributeError: 'A' object attribute 'x' is read-only

    Also note that class variables need to be assigned to a value during definition, and
    this value will not be type checked by `Corgy`.

    Inheritance works as expected, whether base classes are themselves `Corgy` classes
    or not, with sub-classes inheriting the attributes of the base class, and overriding
    any redefined attributes::

        >>> class A:
        ...     x: int

        >>> class B(Corgy, A):
        ...     y: float = 1.0
        ...     z: str

        >>> class C(B):
        ...     y: float = 2.0
        ...     z: str
        ...     w: float

        >>> c = C()
        >>> print(c)
        C(x=<unset>, y=2.0, z=<unset>, w=<unset>)

    Tracking of base class attributes can be disabled by setting `corgy_track_bases` to
    `False` in the class definition. Properties will still be inherited following
    standard inheritance rules, but `Corgy` will ignore them::

        >>> class A:
        ...     x: int

        >>> class B(Corgy, A, corgy_track_bases=False):
        ...     y: float = 1.0
        ...     z: str

        >>> b = B()
        >>> print(b)
        B(y=1.0, z=<unset>)

    `Corgy` instances can be frozen (preventing any further changes) using the `freeze`
    method. This method can be called automatically after `__init__` by by setting
    `corgy_freeze_after_init` to `True` in the class definition::

        >>> class A(Corgy, corgy_freeze_after_init=True):
        ...    x: int

        >>> a = A(x=1)
        >>> a.x = 2
        Traceback (most recent call last):
            ...
        TypeError: cannot set `x`: object is frozen

    `Corgy` recognizes a number of special annotations, which are used to control how
    attribute values are processed.

    Note:
        If any of the following annotations are unavailable in the Python version being
        used, you can import them from `typing_extensions` (which is available on PyPI).

    *Annotations*
    `typing.Annotated` can be used to add additional metadata to attributes, akin to
    doc strings. It is primarily used to control how attributes are added to
    `ArgumentParser` instances. `typing.Annotated` is stripped on class creation,
    leaving only the base type::

        >>> import sys
        >>> if sys.version_info >= (3, 9):
        ...     from typing import Annotated, Literal
        ... else:
        ...     from typing_extensions import Annotated, Literal

        >>> class A(Corgy):
        ...     x: Annotated[int, "this is x"]

        >>> A.attrs()
        {'x': <class 'int'>}

    `Annotated` should always be the outermost type annotation for an attribute.
    Refer to the docs for `Corgy.add_args_to_parser` for details on usage.

    *Required/NotRequired*
    By default, `Corgy` attributes are not required, and can be unset. This can be
    changed by setting `corgy_required_by_default` to `True` in the class definition::

        >>> class A(Corgy, corgy_required_by_default=True):
        ...     x: int

        >>> A()
        Traceback (most recent call last):
            ...
        ValueError: missing required attribute: `x`
        >>> a = A(x=1)
        >>> del a.x
        Traceback (most recent call last):
            ...
        TypeError: attribute `x` cannot be unset

    Attributes can also explicitly be marked as required/not-required using
    `corgy.Required` and `corgy.NotRequired` annotations::

        >>> from corgy import Required, NotRequired

        >>> class A(Corgy):
        ...     x: Required[int]
        ...     y: NotRequired[int]
        ...     z: int  # not required by default

        >>> a = A(x=1)
        >>> print(a)
        A(x=1, y=<unset>, z=<unset>)

        >>> class B(Corgy, corgy_required_by_default=True):
        ...     x: Required[int]
        ...     y: NotRequired[int]
        ...     z: int

        >>> b = B(x=1, z=2)
        >>> print(b)
        B(x=1, y=<unset>, z=2)

    *Optional*
    Annotating an attribute with `typing.Optional` allows it to be `None`::

        >>> from typing import Optional

        >>> class A(Corgy):
        ...     x: Optional[int]

        >>> a = A()
        >>> a.x = None

    In Python >= 3.10, instead of using `typing.Annotated`, `| None` can be used, i.e.,
    `x: int | None` for example.

    Note that `Optional` is not the same as `NotRequired`. `Optional` allows an
    attribute to be `None`, while `NotRequired` allows an attribute to be unset.
    A `Required` `Optional` attribute will need a value (which can be `None`)::

        >>> class A(Corgy):
        ...     x: Required[Optional[int]]

        >>> A()
        Traceback (most recent call last):
            ...
        ValueError: missing required attribute: `x`
        >>> a = A(x=None)
        >>> print(a)
        A(x=None)

    *Collections*
    Several collection types can be used to annotate attributes, which will restrict the
    type of accepted values. Values in the collection will be checked to ensure that
    they match the annotated collection types. The following collection types are
    supported:

    1. `collections.abc.Sequence` (`typing.Sequence` on Python < 3.9)
    2. `tuple` (`typing.Tuple` on Python < 3.9)
    3. `list` (`typing.List` on Python < 3.9)
    4. `set` (`typing.Set` on Python < 3.9)

    There are a few different ways to use these types, each resulting in different
    validation conditions. The simplest case is a plain (possibly empty) collection of a
    single type::

        >>> from typing import List, Sequence, Set, Tuple

        >>> class A(Corgy):
        ...     x: Sequence[int]
        ...     y: Tuple[str]
        ...     z: Set[float]
        ...     w: List[int]

        >>> a = A()
        >>> a.x = [1, 2]
        >>> a.y = ("1", "2")
        >>> a.z = {1.0, 2.0}
        >>> a.w = [1, 2]
        >>> a
        A(x=[1, 2], y=('1', '2'), z={1.0, 2.0}, w=[1, 2])

        >>> a.x = [1, "2"]
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type '<class 'int'>': '2'

        >>> a.x = (1, 2)      # `Sequence` accepts any sequence type

        >>> # `Tuple` only accepts tuples
        >>> a.y = ["1", "2"]  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: error setting `y`: invalid value for type 'typing.Tuple[str]':
        ['1', '2']

    The collection length can be controlled by the arguments of the type annotation.
    Note, however, that `typing.Sequence/typing.List/typing.Set` do not
    accept multiple arguments, and so, cannot be used if collection length has to be
    specified. On Python < 3.9, only `typing.Tuple` can be used for controlling
    collection lengths.

    To specify that a collection must be non-empty, use ellipsis (`...`) as the second
    argument of the type::

        >>> class A(Corgy):
        ...     x: Tuple[int, ...]

        >>> a = A()
        >>> a.x = tuple()  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: expected non-empty collection for type
        'typing.Tuple[int, ...]'

    Collections can also be restricted to be of a fixed length::

        >>> class A(Corgy):
        ...     x: Tuple[int, str]
        ...     y: Tuple[int, int, int]

        >>> a = A()
        >>> a.x = (1, 1)
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type '<class 'str'>': 1

        >>> a.y = (1, 1)  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: error setting `y`: invalid value for type
        'typing.Tuple[int, int, int]': (1, 1): expected exactly '3' elements

    *Literals*
    `typing.Literal` can be used to specify that an attribute takes one of a fixed set
    of values::

        >>> class A(Corgy):
        ...     x: Literal[0, 1, "2"]

        >>> a = A()
        >>> a.x = 0
        >>> a.x = "2"
        >>> a.x = "1"  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type
        'typing.Literal[0, 1, '2']': '1'

    Type annotations can be nested; for instance,
    `Sequence[Literal[0, 1, 2], Literal[0, 1, 2]]` represents a sequence of length 2,
    where each element is either 0, 1, or 2.

    A fixed set of attribute values can also be specified by adding a `__choices__`
    attribute to the argument type, containing a collection of choices::

        >>> class T(int):
        ...     __choices__ = (1, 2)

        >>> class A(Corgy):
        ...     x: T

        >>> a = A()
        >>> a.x = 1
        >>> a.x = 3  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: error setting `x`: invalid value for type '<class 'T'>': 3:
        expected one of: (1, 2)

    Note that choices specified in this way are not type-checked to ensure that they
    match the argument type; in the above example, `__choices__` could be set to
    `(1, "2")`.

    *Self*
    `Corgy` classes can have attributes of their own type, annotated using
    `typing.Self`.

        >>> if sys.version_info >= (3, 11):
        ...     from typing import Self
        ... else:
        ...     from typing_extensions import Self

        >>> class C(Corgy):
        ...     x: int
        ...     c: Self

        >>> c = C(x=1)
        >>> c.c = C(x=2)
        >>> c
        C(x=1, c=C(x=2))

        >>> class D(C):
        ...     ...

        >>> c.c = D(x=3)  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: error setting `c`: invalid value for type 'Self (bound to
        <class 'C'>)': D(x=3)

    """

    @classmethod
    def add_args_to_parser(
        cls,
        parser: _ActionsContainer,
        name_prefix: str = "",
        flatten_subgrps: bool = False,
        defaults: Optional[Mapping[str, Any]] = None,
    ):
        """Add the class' `Corgy` attributes to the given parser.

        Args:
            parser: Argument parser/group to which the attributes will be added.
            name_prefix: Prefix for argument names. Arguments will be named
                `--<name-prefix>:<attr-name>`. If custom flags are present,
                `--<name-prefix>:<flag>` will be used instead (one for each flag).
            flatten_subgrps: Whether to add sub-groups to the main parser instead of
                creating argument groups. Note: sub-sub-groups are always added with
                this argument set to `True`, since `argparse` in unable to properly
                display nested group arguments.
            defaults: Optional mapping with default values for arguments. Any value
                specified here will override default values specified in the class.
                Values for groups can be specified either as `Corgy` instances, or as
                individual values using the same syntax as for `Corgy.from_dict`.

        Type annotations control how attributes are added to the parser. A number of
        special annotations are parsed and stripped from attribute types to determine
        the parameters for calling `ArgumentParser.add_argument`. These special
        annotations are described below.

        Note: `add_args_to_parser` cannot be used if the type annotation for any
        attribute of the class includes `Self`, unless a custom parser is defined
        for such attributes. See docs for `corgyparser` on how to define custom
        parsers.

        *Annotated*
        `typing.Annotated` can be used to add a help message for the argument::

            >>> import argparse
            >>> from argparse import ArgumentParser
            >>> from corgy import CorgyHelpFormatter

            >>> class A(Corgy):
            ...     x: Annotated[int, "help for x"]

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            options:
              --x int  help for x (optional)

        This annotation can also be used to modify the parser flags for the argument. By
        default, the attribute name is used, prefixed with `--`, and with `_` replaced
        by `-`. If the custom flag does not have a leading `-`, a positional argument
        will be created::

            >>> class A(Corgy):
            ...     x: Annotated[int, "help for x", ["-x", "--ex"]]
            ...     y: Annotated[int, "help for y", ["y"]]

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            positional arguments:
              y int        help for y
            <BLANKLINE>
            options:
              -x/--ex int  help for x (optional)

        `Annotated` can accept multiple arguments, but only the first three are used
        by `Corgy`. The first argument is the attribute type, the second is the help
        message (which must be a string), and the third is a sequence of flags.

        *Required/NotRequired*
        Every corgy attribute is either required or not required. The default status
        depends on the class parameter `corgy_required_by_default` (`False` by default).
        Attributes can also be explicitly marked as required or not required, and will
        control whether the argument will be added with `required=True`::

            >>> from corgy import Required, NotRequired

            >>> class A(Corgy):
            ...     x: Required[int]
            ...     y: NotRequired[int]
            ...     z: int

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            options:
              --x int  (required)
              --y int  (optional)
              --z int  (optional)

        Attributes which are not required, and don't have a default value are added
        with `default=argparse.SUPPRESS`, and so will not be in the parsed namespace::

            >>> parser.parse_args(["--x", "1", "--y", "2"])
            Namespace(x=1, y=2)

        *Optional*
        Attributes marked with `typing.Optional` are allowed to be `None`. The
        arguments for these attributes can be passed with no values (i.e. `--x`
        instead of `--x=1` or `--x 1`) to indicate that the value should be `None`.

        Note: Attributes with default values are also "optional" in the sense that
        they can be omitted from the command line. However, they are not the same as
        attributes marked with `Optional`, since the former are not allowed to be
        `None`. Furthermore, `Required` `Optional` attributes without default values
        _will_ need to be passed on the command line (possibly with no values).

            >>> class A(Corgy):
            ...     x: Required[Optional[int]]

            >>> parser = ArgumentParser()
            >>> A.add_args_to_parser(parser)
            >>> parser.parse_args(["--x"])
            Namespace(x=None)

        *Boolean*
        `bool` types (when not in a collection) are converted to a pair of options::

            >>> class A(Corgy):
            ...     arg: bool

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            options:
              --arg/--no-arg

        *Collection*
        Collection types are added to the parser by setting `nargs`. The value for
        `nargs` is determined by the collection type. Plain collections, such as
        `Sequence[int]`, will be added with `nargs=*`; Non-empty collections, such as
        `Sequence[int, ...]`, will be added with `nargs=+`; Finally, fixed-length
        collections, such as `Sequence[int, int, int]`, will be added with `nargs` set
        to the length of the collection.

        In all cases, collection types can only be added to a parser if they are single
        type. Heterogenous collections, such as `Sequence[int, str]` cannot be added,
        and will raise `ValueError`. Untyped collections (e.g., `x: Sequence`), also
        cannot be added.

        Arguments for optional collections will also accept no values to indicate
        `None`. Due to this, it is not possible to parse an empty collection for
        an optional collection argument::

            >>> class A(Corgy):
            ...     x: Optional[Sequence[int]]
            ...     y: Sequence[int]

            >>> parser = ArgumentParser()
            >>> A.add_args_to_parser(parser)
            >>> parser.parse_args(["--x", "--y"])
            Namespace(x=None, y=[])

        *Literal*
        For `Literal` types, the provided values are passed to the `choices` argument
        of `ArgumentParser.add_argument`. All values must be of the same type, which
        will be inferred from the type of the first value. If the first value has a
        `__bases__` attribute, the type will be inferred as the first base type, and
        all other choices must be subclasses of that type::

            >>> class T: ...
            >>> class T1(T): ...
            >>> class T2(T): ...

            >>> class A(Corgy):
            ...     x: Literal[T1, T2]

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            options:
              --x T  ({T1/T2} optional)

        For types which specify choices by defining `__choices__`, the values are
        passed to the `choices` argument as with `Literal`, but no type inference is
        performed, and the base attribute type will be used as the argument type.

        **Single-value Literals**
        A special case for `Literal` types is when there is only one choice. In this
        case, the argument is added as a `store_const` action, with the value as the
        `const` argument. A further special case is when the choice is `True/False`,
        in which case the action is `store_true`/`store_false` respectively::

            >>> class A(Corgy):
            ...     x: Literal[True]
            ...     y: Literal[False]
            ...     z: Literal[42]

            >>> parser = ArgumentParser()
            >>> A.add_args_to_parser(parser)
            >>> parser.parse_args(["--x"])  # Note that `y` and `z` are absent
            Namespace(x=True)
            >>> parser.parse_args(["--y"])
            Namespace(y=False)
            >>> parser.parse_args(["--z"])
            Namespace(z=42)

        Note: This special case only applies to `Literal` types, and not types which
        define `__choices__`.

        *Corgy*
        Attributes which are themselves `Corgy` types are treated as argument groups.
        Group arguments are added to the command line parser with the group attribute
        name prefixed. Note that groups will ignore any custom flags when computing the
        prefix; elements within the group will use custom flags, but because they are
        prefixed with `--`, they will not be positional.

        Example::

            >>> class G(Corgy):
            ...     x: int = 0
            ...     y: float

            >>> class A(Corgy):
            ...     x: int
            ...     g: G

            >>> parser = ArgumentParser(
            ...     formatter_class=CorgyHelpFormatter,
            ...     add_help=False,
            ...     usage=argparse.SUPPRESS,
            ... )

            >>> A.add_args_to_parser(parser)
            >>> parser.print_help()
            options:
              --x int      (optional)
            <BLANKLINE>
            g:
              --g:x int    (default: 0)
              --g:y float  (optional)

        **Custom parsers**

        Attributes for which a custom parser is defined using `@corgyparser` will
        be added with a custom action that will call the parser. Refer to the
        documentation for `corgyparser` for details.

        **Metavar**

        This function will not explicitly pass a value for the `metavar` argument of
        `ArgumentParser.add_argument`, unless an attribute's type defines `__metavar__`,
        in which case, it will be passed as is. To change the metavar for attributes
        with custom parsers, set the `metavar` argument of `corgyparser`.
        """
        base_parser = parser
        base_defaults = getattr(cls, "__defaults").copy()
        if defaults is not None:
            base_defaults.update(defaults)

        # Extract default values for group arguments specified individually using
        # the `<group>:<var name>` syntax.
        group_arg_defaults: Dict[str, Dict[str, Any]] = defaultdict(dict)
        for _k, _v in base_defaults.items():
            if ":" in _k:
                _grp_name, _var_name = _k.split(":")
                group_arg_defaults[_grp_name][_var_name] = _v
            elif _k not in cls.__annotations__:
                raise ValueError(f"default value for unknown argument: `{_k}`")

        required_attrs = getattr(cls, "__required")
        custom_flags = getattr(cls, "__flags")
        custom_parsers = getattr(cls, "__parsers")

        @dataclass
        class _Arg:
            name: str
            required: Optional[bool] = None
            positional: Optional[bool] = None
            dest: Optional[str] = None
            flags: Optional[Sequence[str]] = None
            help: Optional[str] = None
            nargs: Union[int, Literal["+", "*"], None] = None
            action: Optional[Type[argparse.Action]] = None
            choices: Optional[Sequence[Any]] = None
            metavar: Optional[str] = None
            add_type: Optional[Any] = None
            const: Optional[Any] = None

            def get_add_kwargs(self) -> Dict[str, Any]:
                assert self.required is not None
                assert self.positional is not None
                assert self.dest is not None
                kwargs: Dict[str, Any] = {}
                if (not self.positional) and (self.name in custom_flags):
                    kwargs["dest"] = self.dest
                if self.help is not None:
                    kwargs["help"] = self.help
                if self.nargs is not None:
                    kwargs["nargs"] = self.nargs
                if self.action is not None:
                    kwargs["action"] = self.action
                if self.choices is not None:
                    kwargs["choices"] = self.choices

                if self.name in base_defaults:
                    kwargs["default"] = base_defaults[self.name]
                elif self.required and not self.positional:
                    kwargs["required"] = True
                elif not self.positional:
                    kwargs["default"] = argparse.SUPPRESS

                if self.metavar is not None:
                    kwargs["metavar"] = self.metavar
                if self.add_type is not None:
                    kwargs["type"] = self.add_type
                if self.const is not None:
                    kwargs["const"] = self.const

                return kwargs

            def process_optional(self, type_) -> Any:
                if is_optional_type(type_):
                    self.action = OptionalTypeAction
                    return type_.__args__[0]
                return type_

            def process_collection(self, type_) -> Any:
                _col_type = get_concrete_collection_type(type_)
                if _col_type is not None:
                    if (
                        not hasattr(type_, "__args__")
                        or not type_.__args__
                        or isinstance(type_.__args__[0], TypeVar)
                    ):
                        raise TypeError(
                            f"`{self.name}` is a collection, but has no type arguments:"
                            f" use `{type_}[<types>]"
                        )
                    if len(type_.__args__) == 1:
                        self.nargs = "*"
                    elif len(type_.__args__) == 2 and type_.__args__[1] is Ellipsis:
                        # `...` is used to represent non-empty collections, e.g.,
                        # `Sequence[int, ...]`.
                        self.nargs = "+"
                    else:
                        # Ensure single type.
                        if any(_a != type_.__args__[0] for _a in type_.__args__[1:]):
                            raise TypeError(
                                f"`{self.name}` has unsupported type `{type_}`: only"
                                f" single-type collections are supported"
                            )
                        self.nargs = len(type_.__args__)
                    return type_.__args__[0]
                return type_

            def process_choices(self, type_) -> Any:
                _is_literal_type = is_literal_type(type_)
                if _is_literal_type:
                    # Determine if the first choice has `__bases__`, in which case
                    # the first base class is the type for the argument.
                    try:
                        c0_type = type_.__args__[0].__bases__[0]
                    except AttributeError:
                        c0_type = type(type_.__args__[0])
                        f_check_type: Callable[[Any, Any], bool] = isinstance
                    else:
                        f_check_type = issubclass

                    # All choices must be of the same type.
                    if any(not f_check_type(_a, c0_type) for _a in type_.__args__[1:]):
                        raise TypeError(
                            f"choices for `{self.name}` not all of type `{c0_type}`: "
                            f"`{type_.__args__}`"
                        )
                    self.choices = type_.__args__

                    # Convert single choice attributes to `store_*` actions.
                    if (
                        self.choices is not None
                        and len(self.choices) == 1
                        and self.nargs is None
                        and self.action is None
                    ):
                        _choice = self.choices[0]
                        if _choice is True:
                            self.action = _StoreTrueAction
                        elif _choice is False:
                            self.action = _StoreFalseAction
                        else:
                            self.action = _StoreConstAction
                            self.const = _choice
                        self.choices = None
                        return None
                    return c0_type

                if hasattr(type_, "__choices__"):
                    self.choices = type_.__choices__

                return type_

        for var_name, var_type in cls.attrs().items():
            var_arg = _Arg(var_name)
            var_arg.help = getattr(
                cls, var_name
            ).__doc__  # doc is stored in the property
            var_arg.required = (
                var_name not in base_defaults and var_name in required_attrs
            )

            # Determine add flags.
            var_arg.flags = custom_flags.get(
                var_name, [f"--{var_name.replace('_', '-')}"]
            )
            assert var_arg.flags is not None
            if name_prefix:
                var_arg.flags = [
                    f"--{name_prefix.replace('_', '-')}:{flag.lstrip('-')}"
                    for flag in var_arg.flags
                ]
                var_arg.dest = f"{name_prefix}:{var_name}"
            else:
                var_arg.dest = var_name

            # Determine if argument is positional.
            if not any(_flag.startswith("-") for _flag in var_arg.flags):
                # Note: the flags cannot be passed to `add_argument` with `dest` set
                # to `var_name` since `argparse` will raise an error for passing `dest`
                # twice (for positional arguments, `argparse` uses the flag to infer the
                # `dest`).
                var_arg.flags = [var_name]
                var_arg.positional = True
            elif all(_flag.startswith("-") for _flag in var_arg.flags):
                var_arg.positional = False
            else:
                raise TypeError(
                    f"inconsistent positional/optional flags for {var_name}: "
                    f"{var_arg.flags}"
                )

            ###################################################################
            # Check if the variable has a custom parser.
            if var_name in custom_parsers:
                _var_parser = custom_parsers[var_name]
                _var_base_type = var_type
                # Extract choices if present.
                if is_literal_type(var_type):
                    _var_choices = var_type.__args__
                    try:
                        _var_base_type = _var_choices[0].__bases__[0]
                    except AttributeError:
                        _var_base_type = type(_var_choices[0])
                elif hasattr(var_type, "__choices__"):
                    _var_choices = var_type.__choices__
                else:
                    _var_choices = None

                var_arg.action = partial(
                    CorgyParserAction, _var_parser, _var_choices  # type: ignore
                )
                var_arg.add_type = str
                var_arg.nargs = getattr(_var_parser, "__nargs__", None)
                try:
                    var_arg.metavar = _var_parser.__metavar__
                except AttributeError:
                    try:
                        var_arg.metavar = _var_base_type.__metavar__
                    except AttributeError:
                        pass
                parser.add_argument(*var_arg.flags, **var_arg.get_add_kwargs())
                continue

            ###################################################################
            # Check if the variable is also `Corgy` type.
            if type(var_type) is type(cls):
                # Create an argument group using `<var_type>`.
                # If there is a default value, pass it using `**defaults`.
                if var_name in base_defaults:
                    try:
                        grp_defaults = base_defaults[var_name].as_dict()
                    except AttributeError:
                        raise ValueError(
                            f"default value for `{var_name}` is not a `Corgy` instance"
                        ) from None
                else:
                    grp_defaults = {}

                # Update defaults with any values specified individually.
                grp_defaults.update(group_arg_defaults.get(var_name, {}))

                grp_parser: _ActionsContainer
                if flatten_subgrps:
                    grp_parser = base_parser
                else:
                    grp_parser = base_parser.add_argument_group(
                        var_arg.dest, var_arg.help
                    )
                var_type.add_args_to_parser(
                    grp_parser, var_arg.dest, True, grp_defaults
                )
                continue

            ###################################################################
            # Check if the variable is boolean. Boolean variables are converted to
            # `--<var-name>`/`--no-<var-name>` arguments (if not positional).
            if var_type is bool:
                var_arg.add_type = bool
                if not var_arg.positional:
                    var_arg.action = BooleanOptionalAction
                parser.add_argument(*var_arg.flags, **var_arg.get_add_kwargs())
                continue

            ###################################################################
            # Process annotations.
            var_base_type = var_type
            var_base_type = var_arg.process_optional(var_base_type)
            var_base_type = var_arg.process_collection(var_base_type)
            var_base_type = var_arg.process_choices(var_base_type)

            if var_base_type is Self:
                raise TypeError(
                    "'add_args_to_parser' cannot be used with 'Self' type present"
                )

            var_arg.metavar = getattr(var_base_type, "__metavar__", None)
            var_arg.add_type = var_base_type
            parser.add_argument(*var_arg.flags, **var_arg.get_add_kwargs())

    def __init__(self, **args):
        if self.__class__ is Corgy:
            raise TypeError("`Corgy` is an abstract class and cannot be instantiated")

        setattr(self, f"_{self.__class__.__name__.lstrip('_')}__frozen", False)

        cls_attrs = self.attrs()
        cls_defaults = getattr(self, "__defaults")
        for attr_name in cls_attrs:
            if attr_name in args:
                setattr(self, attr_name, args[attr_name])
            elif attr_name in cls_defaults:
                setattr(self, attr_name, cls_defaults[attr_name])
            elif attr_name in getattr(self, "__required"):
                raise ValueError(f"missing required attribute: `{attr_name}`")

        if getattr(self.__class__, "__freeze_after_init"):
            self.freeze()

    def _str(self, f_str: Callable[..., str]) -> str:
        s = f"{self.__class__.__name__}("
        i = 0
        for arg_name in self.attrs():
            try:
                _val_s = f_str(getattr(self, arg_name))
            except AttributeError:
                if f_str is repr:
                    continue
                _val_s = "<unset>"
            if i != 0:
                s = s + ", "
            s = s + f"{arg_name}="
            s = s + _val_s
            i += 1
        s = s + ")"
        return s

    def __repr__(self) -> str:
        return self._str(repr)

    def __str__(self) -> str:
        return self._str(str)

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        if self.__class__ is not other.__class__:
            return False
        for _attr in self.attrs():
            _self_has, _other_has = hasattr(self, _attr), hasattr(other, _attr)
            if _self_has != _other_has:
                # One instance has `_attr` set; the other doesn't.
                return False
            if not _self_has:
                # Both instances don't have `_attr` set.
                continue
            if getattr(self, _attr) != getattr(other, _attr):
                return False
        return True

    @classmethod
    def attrs(cls) -> Dict[str, Type]:
        """Return a dictionary mapping attributes of the class to their types.

        Example::

            >>> class A(Corgy):
            ...     x: Annotated[int, "x"]
            ...     y: Sequence[str]

            >>> A.attrs()
            {'x': <class 'int'>, 'y': typing.Sequence[str]}

        """
        return {
            _attr: getattr(cls, _attr).fget.__annotations__["return"]
            for _attr in cls.__annotations__
        }

    def as_dict(self, recursive: bool = True, flatten: bool = False) -> Dict[str, Any]:
        """Return the object as a dictionary.

        The returned dictionary maps attribute names to their values. Unset attributes
        are omitted, unless they have default values.

        Args:
            recursive: whether to recursively call `as_dict` on attributes which are
                `Corgy` instances. Otherwise, they are returned as is.
            flatten: whether to flatten group arguments into `:` separated strings.
                Only takes effect if `recursive` is `True`.

        Examples::

            >>> class G(Corgy):
            ...     x: int

            >>> g = G(x=1)
            >>> g.as_dict()
            {'x': 1}

            >>> class A(Corgy):
            ...     x: str
            ...     g: G

            >>> a = A(x="one", g=g)
            >>> a.as_dict(recursive=False)
            {'x': 'one', 'g': G(x=1)}
            >>> a.as_dict()
            {'x': 'one', 'g': {'x': 1}}
            >>> a.as_dict(flatten=True)
            {'x': 'one', 'g:x': 1}

        """

        def dictify_corgys(_val):
            _coll_type = get_concrete_collection_type(type(_val))
            if _coll_type is not None:
                _cast_type = _coll_type if _coll_type is not AbstractSequence else list
                return _cast_type(  # pylint: disable=abstract-class-instantiated
                    [dictify_corgys(_val_part) for _val_part in _val]
                )
            if isinstance(_val.__class__, CorgyMeta):
                return _val.as_dict(recursive=True, flatten=flatten)
            return _val

        self_dict = {}
        for attr_name, attr_type in self.attrs().items():
            try:
                attr_val = getattr(self, attr_name)
            except AttributeError:
                continue

            if recursive:
                attr_val = dictify_corgys(attr_val)
                if flatten and (isinstance(attr_type, CorgyMeta) or attr_type is Self):
                    for _k, _v in attr_val.items():
                        _flat_key = f"{attr_name}:{_k}"
                        self_dict[_flat_key] = _v
                    continue

            self_dict[attr_name] = attr_val

        return self_dict

    @classmethod
    def from_dict(cls: Type[_T], d: Mapping[str, Any], try_cast: bool = False) -> _T:
        """Return a new instance of the class using a dictionary.

        This is roughly equivalent to `cls(**d)`, with the main exception being that
        groups can be specified as dictionaries themselves, and will be processed
        recursively.

        Args:
            d: Dictionary to create the instance from.
            try_cast: Whether to try and cast values which don't match attribute types.

        Example::

            >>> class G(Corgy):
            ...     x: int

            >>> class A(Corgy):
            ...     x: str
            ...     g: G

            >>> A.from_dict({"x": "one", "g": G(x=1)})
            A(x='one', g=G(x=1))
            >>> A.from_dict({"x": "one", "g": {"x": 1}})
            A(x='one', g=G(x=1))
            >>> A.from_dict({"x": "one", "g": {"x": "1"}}, try_cast=True)
            A(x='one', g=G(x=1))
            >>> G.from_dict({"x": "1"})
            Traceback (most recent call last):
                ...
            ValueError: error setting `x`: invalid value for type '<class 'int'>': '1'

        Group attributes can also be passed directly in the dictionary by prefixing
        their names with the group name and a colon::

            >>> A.from_dict({"x": "one", "g:x": 1})
            A(x='one', g=G(x=1))

            >>> class B(Corgy):
            ...     x: float
            ...     a: A

            >>> B.from_dict({"x": 1.1, "a:x": "one", "a:g:x": 1})
            B(x=1.1, a=A(x='one', g=G(x=1)))

        """
        main_args_map = {}
        grp_args_map: Dict[str, Any] = defaultdict(dict)
        cls_attrs = cls.attrs()

        for arg_name, arg_val in d.items():
            if ":" in arg_name:
                grp_name, arg_name_base = arg_name.split(":", maxsplit=1)
                if not hasattr(cls, grp_name):
                    raise ValueError(
                        f"invalid argument `{arg_name}`: "
                        f"`{cls}` has no group named `{grp_name}`"
                    )
                if grp_name in d:
                    raise ValueError(
                        f"conflicting arguments: `{arg_name}` and `{grp_name}`"
                    )
                grp_type = cls_attrs[grp_name]
                if not isinstance(grp_type, CorgyMeta):
                    raise ValueError(f"`{grp_name}` is not a `Corgy` class")
                grp_args_map[grp_name][arg_name_base] = arg_val

            elif hasattr(cls, arg_name):
                arg_type = cls_attrs[arg_name]
                if isinstance(arg_type, CorgyMeta) and isinstance(arg_val, dict):
                    grp_args_map[arg_name] = arg_val
                else:
                    main_args_map[arg_name] = arg_val

        for grp_name, grp_args in grp_args_map.items():
            grp_type = cls_attrs[grp_name]
            main_args_map[grp_name] = grp_type.from_dict(grp_args, try_cast)

        cls_attrs = cls.attrs()
        for arg_name, arg_val in main_args_map.copy().items():
            if arg_name in cls_attrs:
                try:
                    main_args_map[arg_name] = check_val_type(
                        arg_val,
                        cls_attrs[arg_name],
                        try_cast,
                        try_load_corgy_dicts=True,
                        self_type=cls,
                    )
                except ValueError as e:
                    raise ValueError(f"error setting `{arg_name}`: {e}") from None
        return cls(**main_args_map)

    def load_dict(
        self, d: Dict[str, Any], try_cast: bool = False, strict: bool = False
    ) -> None:
        """Load a dictionary into an instance of the class.

        Previous attributes are overwritten. Sub-dictionaries will be parsed
        recursively if the corresponding attribute already exists, else will be parsed
        using `from_dict`. As with `from_dict`, items in the dictionary without
        corresponding attributes are ignored.

        Args:
            d: Dictionary to load.
            try_cast: Whether to try and cast values which don't match attribute types.
            strict: If `True`, attributes with existing values that are not in the
                dictionary will be unset.

        Example::

            >>> class A(Corgy):
            ...     x: int
            ...     y: str
            >>> a = A(x=1)
            >>> _i = id(a)
            >>> a.load_dict({"y": "two"})
            >>> a
            A(x=1, y='two')
            >>> _i == id(a)
            True
            >>> a.load_dict({"y": "three"}, strict=True)
            >>> a
            A(y='three')
            >>> _i == id(a)
            True
            >>> a = A()
            >>> a.load_dict({"x": "1"})
            Traceback (most recent call last):
                ...
            ValueError: error setting `x`: invalid value for type '<class 'int'>': '1'
            >>> a.load_dict({"x": "1"}, try_cast=True)
            >>> a
            A(x=1)

        """
        main_args_map: Dict[str, Any] = defaultdict(dict)
        cls_attrs = self.attrs()

        for arg_name, arg_val in d.items():
            if ":" in arg_name:
                grp_name, arg_name_base = arg_name.split(":", maxsplit=1)
                if grp_name not in cls_attrs:
                    raise ValueError(
                        f"invalid argument `{arg_name}`: "
                        f"`{self.__class__}` has no group named `{grp_name}`"
                    )
                if grp_name in d:
                    raise ValueError(
                        f"conflicting arguments: `{arg_name}` and `{grp_name}`"
                    )
                grp_type = cls_attrs[grp_name]
                if not isinstance(grp_type, CorgyMeta):
                    raise ValueError(f"`{grp_name}` is not a `Corgy` class")
                main_args_map[grp_name][arg_name_base] = arg_val

            elif arg_name in cls_attrs:
                main_args_map[arg_name] = arg_val

        for arg_name, arg_type in cls_attrs.items():
            if arg_name not in main_args_map:
                if strict and hasattr(self, arg_name):
                    delattr(self, arg_name)
                continue

            arg_new_val = main_args_map[arg_name]
            if isinstance(arg_type, CorgyMeta) and isinstance(arg_new_val, dict):
                try:
                    arg_obj = getattr(self, arg_name)
                except AttributeError:
                    setattr(self, arg_name, arg_type.from_dict(arg_new_val))
                else:
                    arg_obj.load_dict(arg_new_val, try_cast, strict)
            else:
                try:
                    arg_new_val = check_val_type(
                        arg_new_val,
                        arg_type,
                        try_cast,
                        try_load_corgy_dicts=True,
                        self_type=type(self),
                    )
                except ValueError as e:
                    raise ValueError(f"error setting `{arg_name}`: {e}") from None
                setattr(self, arg_name, arg_new_val)

    @classmethod
    def parse_from_cmdline(
        cls: Type[_T],
        parser: Optional[ArgumentParser] = None,
        defaults: Optional[Mapping[str, Any]] = None,
        **parser_args,
    ) -> _T:
        """Return an instance of the class parsed from command line arguments.

        Args:
            parser: An instance of `argparse.ArgumentParser` or `None`. If `None`, a new
                instance is created.
            defaults: A dictionary of default values for the attributes, passed to
                `add_args_to_parser`. Refer to the docs for `add_args_to_parser` for
                more details.
            parser_args: Arguments to be passed to `argparse.ArgumentParser()`. Ignored
                if `parser` is not None.

        Raises:
            ArgumentTypeError: Error parsing command line arguments.
        """
        if parser is None:
            if "formatter_class" not in parser_args:
                parser_args["formatter_class"] = CorgyHelpFormatter
            parser = ArgumentParser(**parser_args)
        cls.add_args_to_parser(parser, defaults=defaults)
        args = vars(parser.parse_args())
        return cls.from_dict(args, try_cast=True)

    @classmethod
    def parse_from_toml(
        cls: Type[_T],
        toml_file: IO[bytes],
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> _T:
        """Parse an object of the class from a toml file.

        Args:
            toml_file: A file-like object containing the class attributes in toml.
            defaults: A dictionary of default values, overriding any values specified
                in the class.

        Raises:
            TOMLDecodeError: Error parsing the toml file.

        Example::

            >>> class G(Corgy):
            ...     x: int
            ...     y: Sequence[int]

            >>> class A(Corgy):
            ...     x: str
            ...     g: G

            >>> from io import BytesIO
            >>> f = BytesIO(b'''
            ...     x = 'one'
            ...     [g]
            ...     x = 1
            ...     y = [1, 2, 3]
            ... ''')

            >>> A.parse_from_toml(f)  # doctest: +SKIP
            A(x='one', g=G(x=1, y=[1, 2, 3]))

        """
        tomli = import_module("tomllib" if sys.version_info >= (3, 11) else "tomli")
        toml_data = tomli.load(toml_file)
        if defaults is not None:
            for _k, _v in defaults.items():
                if _k not in toml_data:
                    toml_data[_k] = _v
        _parsers = getattr(cls, "__parsers")
        for _k, _v in toml_data.items():
            if _k in _parsers:
                toml_data[_k] = _parsers[_k](_v)
        return cls.from_dict(toml_data, try_cast=True)

    def freeze(self):
        """Freeze the object, preventing any further changes to attributes.

        Example::

            >>> class A(Corgy):
            ...     x: int
            ...     y: int

            >>> a = A(x=1, y=2)
            >>> a.x = 2
            >>> a.freeze()
            >>> a.x = 3
            Traceback (most recent call last):
                ...
            TypeError: cannot set `x`: object is frozen
            >>> del a.y
            Traceback (most recent call last):
                ...
            TypeError: cannot delete `y`: object is frozen

        """
        setattr(self, f"_{self.__class__.__name__.lstrip('_')}__frozen", True)



import argparse
import sys
from argparse import ArgumentParser, ArgumentTypeError
from contextlib import redirect_stderr
from functools import partial
from io import StringIO
from typing import ClassVar, Optional, Tuple
from unittest import TestCase
from unittest.mock import MagicMock, patch

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if sys.version_info >= (3, 9):
    from typing import Annotated, Literal
else:
    from typing_extensions import Annotated, Literal

from corgy import Corgy, corgyparser, Required
from corgy._corgyparser import CorgyParserAction


class TestCorgyCustomParsers(TestCase):
    def test_corgyparser_raises_if_not_passed_name(self):
        with self.assertRaises(TypeError):

            @corgyparser
            def spam():
                ...

    def test_corgy_raises_if_corgyparser_target_invalid(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int

                @corgyparser("y")
                def parsex(s):  # type: ignore # pylint: disable=no-self-argument
                    return 0

    def test_corgy_raises_if_corgyparser_target_not_annotated(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int
                y = 1

                @corgyparser("y")
                @staticmethod
                def parsex(s):
                    return 0

    def test_corgy_raises_if_corgyparser_target_classvar(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: ClassVar[int]

                @corgyparser("x")
                @staticmethod
                def parsex(s):
                    return 0

    def test_corgyparser_raises_on_nargs_mismatch(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int
                y: int

                @corgyparser("x")
                @corgyparser("y", nargs="+")
                @staticmethod
                def parsex(s):
                    return 0

        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int
                y: int

                @corgyparser("x", nargs=2)
                @corgyparser("y", nargs="*")
                @staticmethod
                def parsex(s):
                    return 0

    def test_add_args_handles_corgyparser(self):
        class C(Corgy):
            x: Annotated[int, "x"]

            @corgyparser("x")
            def parsex(s):  # type: ignore # pylint: disable=no-self-argument
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x", type=str, help="x", action=_parser_action, default=argparse.SUPPRESS
        )

    def test_add_args_with_custom_parser_respects_default_value(self):
        class C(Corgy):
            x: int = 1

            @corgyparser("x")
            def parsex(s):  # type: ignore # pylint: disable=no-self-argument
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x", type=str, default=1, action=_parser_action
        )

    def test_cmdline_parsing_calls_custom_parser(self):
        class C(Corgy):
            x: int

            @corgyparser("x")
            def parsex(s):  # type: ignore # pylint: disable=no-self-argument
                return 0

        getattr(C, "__parsers")["x"] = MagicMock(return_value=0)
        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "test"])

        C.parse_from_cmdline(parser)
        getattr(C, "__parsers")["x"].assert_called_once_with("test")

    def test_cmdline_parsing_calls_custom_parser_with_specified_nargs(self):
        orig_parse_args = ArgumentParser.parse_args

        def _run_and_check(cls, nargs, cmd_args, expected_call_args):
            getattr(cls, "__parsers")["x"] = MagicMock(return_value=0, __nargs__=nargs)
            parser = ArgumentParser()
            parser.parse_args = lambda: orig_parse_args(parser, ["--x"] + cmd_args)
            parser.error = MagicMock(side_effect=ArgumentTypeError)
            cls.parse_from_cmdline(parser)
            getattr(cls, "__parsers")["x"].assert_called_once_with(*expected_call_args)

        for nargs in [None, "*", "+", 3]:

            class C(Corgy):
                x: int

                @corgyparser("x", nargs=nargs)  # pylint: disable=cell-var-from-loop
                @staticmethod
                def parsex(s):
                    return 0

            if nargs is None:
                _run_and_check(C, nargs, ["x"], ["x"])
                # _run_and_check(C, [], [])
                with self.assertRaises(ArgumentTypeError):
                    _run_and_check(C, nargs, ["x", "y"], [["x", "y"]])
            elif nargs == "*":
                _run_and_check(C, nargs, ["x"], [["x"]])
                _run_and_check(C, nargs, [], [[]])
            elif nargs == "+":
                _run_and_check(C, nargs, ["x"], [["x"]])
                with self.assertRaises(ArgumentTypeError):
                    _run_and_check(C, nargs, [], [])
            else:
                _run_and_check(C, nargs, ["x", "y", "z"], [["x", "y", "z"]])
                with self.assertRaises(ArgumentTypeError):
                    _run_and_check(C, nargs, ["x", "y"], [["x", "y"]])

    def test_cmdline_parsing_returns_custom_parser_output(self):
        class C(Corgy):
            x: int

            @corgyparser("x")
            def parsex(s):  # type: ignore # pylint: disable=no-self-argument
                return -1

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "test"])

        args = C.parse_from_cmdline(parser)
        self.assertEqual(args.x, -1)

    def test_corgyparser_allows_decorating_staticmethod(self):
        class C(Corgy):
            x: int

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return 0

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "test"])

        c = C.parse_from_cmdline(parser)
        self.assertEqual(c.x, 0)

    def test_corgyparser_raises_if_decorating_non_staticmethod(self):
        with self.assertRaises(TypeError):

            class _(Corgy):
                x: int

                @corgyparser("x")
                @classmethod
                def parsex(cls, s):
                    return 0

    def test_corgyparser_functions_are_callable(self):
        class C(Corgy):
            x: int
            y: int

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return 0

            @corgyparser("y")
            def parsey(s):  # type: ignore # pylint: disable=no-self-argument
                return 1

        self.assertEqual(C.parsex("x"), 0)
        self.assertEqual(C.parsey("y"), 1)

    def test_corgyparser_accepts_multiple_arguments(self):
        class C(Corgy):
            x: int
            y: int

            @corgyparser("x", "y")
            @staticmethod
            def parsexy(s):
                return 0

        self.assertIs(getattr(C, "__parsers")["x"], getattr(C, "__parsers")["y"])

    def test_corgyparser_decorators_can_be_chained(self):
        class C(Corgy):
            x: int
            y: int

            @corgyparser("x")
            @corgyparser("y")
            @staticmethod
            def parsexy(s):
                return 0

        self.assertIs(getattr(C, "__parsers")["x"], getattr(C, "__parsers")["y"])

    def test_corgyparser_allows_setting_metavar(self):
        class C(Corgy):
            x: int

            @corgyparser("x", metavar="custom")
            @staticmethod
            def parsex(s):
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x",
            type=str,
            action=_parser_action,
            default=argparse.SUPPRESS,
            metavar="custom",
        )

    def test_corgyparser_handles_setting_metavar_with_chaining(self):
        class C(Corgy):
            x: int
            y: int
            z: int
            w: int

            @corgyparser("x")
            @corgyparser("y", metavar="custom y")
            @corgyparser("z")
            @corgyparser("w", metavar="custom w")
            @staticmethod
            def parsexyzw(s):
                return 0

        self.assertEqual(C.parsexyzw.fparse.__metavar__, "custom y")

    def test_add_args_with_custom_parser_uses_custom_metavar(self):
        class T:
            __metavar__ = "T"

        class C(Corgy):
            x: T

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x",
            type=str,
            action=_parser_action,
            default=argparse.SUPPRESS,
            metavar="T",
        )

    def test_add_args_handles_bool_with_custom_parser(self):
        class C(Corgy):
            x: bool

            @corgyparser("x", nargs=2, metavar=("a", "b"))
            @staticmethod
            def parsex(s):
                return s[0] == s[1]

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x",
            type=str,
            action=_parser_action,
            default=argparse.SUPPRESS,
            nargs=2,
            metavar=("a", "b"),
        )

    def test_corgyparser_metavar_overrides_type_metavar(self):
        class T:
            __metavar__ = "T"

        class C(Corgy):
            x: T

            @corgyparser("x", metavar="custom")
            @staticmethod
            def parsex(s):
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x",
            type=str,
            action=_parser_action,
            default=argparse.SUPPRESS,
            metavar="custom",
        )

    def test_corgy_cls_inherits_custom_parser(self):
        class C(Corgy):
            x: int

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return int(s[0]) + 1

        class D(C):
            ...

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "1"])

        d = D.parse_from_cmdline(parser)
        self.assertEqual(d.x, 2)

    def test_corgy_cls_overrides_nargs_with_custom_parser(self):
        class C1(Corgy):
            x: Tuple[int]

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return 0

        class C2(C1):
            x: Tuple[int, ...]  # type: ignore

        class C3(C1):
            x: Tuple[int, int, int]  # type: ignore

        for C in (C1, C2, C3):
            with self.subTest(cls=C.__name__):
                parser = ArgumentParser()
                parser.add_argument = MagicMock()
                _parser_action = partial(CorgyParserAction, C.parsex)
                with patch(
                    "corgy._corgy.partial", MagicMock(return_value=_parser_action)
                ):
                    C.add_args_to_parser(parser)
                parser.add_argument.assert_called_once_with(
                    "--x", type=str, default=argparse.SUPPRESS, action=_parser_action
                )

    def test_corgy_cls_respects_choices_with_custom_parser(self):
        class C(Corgy):
            x: Literal[10, 20, 30]

            @corgyparser("x", nargs="+")
            @staticmethod
            def parsex(s):
                return sum(map(int, s))

        orig_parse_args = ArgumentParser.parse_args

        def _run_with_args(*cmd_args):
            with redirect_stderr(StringIO()):
                parser = ArgumentParser()
                parser.parse_args = lambda: orig_parse_args(
                    parser, ["--x"] + list(cmd_args)
                )
                args = C.parse_from_cmdline(parser)
                return args.x

        with self.assertRaises(SystemExit):
            _run_with_args("2", "3", "4")
        self.assertEqual(_run_with_args("1", "2", "3", "4"), 10)

    def test_corgy_cls_respects_optional_with_custom_parser(self):
        class C(Corgy):
            x: Optional[int]

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return 0

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x", type=str, action=_parser_action, default=argparse.SUPPRESS
        )

    def test_cmdline_parsing_of_complex_nested_types_works_with_custom_parser(self):
        class C(Corgy):
            x: Tuple[Tuple[int, float], ...]

            @corgyparser("x", nargs="*")
            @staticmethod
            def parsex(s_tup):
                if not s_tup:
                    raise ValueError
                if len(s_tup) % 2:
                    raise ValueError
                o_list = []
                for _i in range(len(s_tup) // 2):
                    s = [s_tup[2 * _i], s_tup[2 * _i + 1]]
                    o_list.append((int(s[0]), float(s[1])))
                return tuple(o_list)

        orig_parse_args = ArgumentParser.parse_args

        def _run_with_args(*cmd_args):
            with redirect_stderr(StringIO()):
                parser = ArgumentParser()
                parser.parse_args = lambda: orig_parse_args(
                    parser, ["--x"] + list(cmd_args)
                )
                args = C.parse_from_cmdline(parser)
                return args.x

        self.assertTupleEqual(_run_with_args("1", "2.1"), ((1, 2.1),))
        self.assertTupleEqual(
            _run_with_args("1", "2.1", "3", "4.1"), ((1, 2.1), (3, 4.1))
        )
        with self.assertRaises(SystemExit):
            _run_with_args("1", "two")
        with self.assertRaises(SystemExit):
            _run_with_args("1.1", "2.1")
        with self.assertRaises(SystemExit):
            _run_with_args("1", "2.1", "3")
        with self.assertRaises(SystemExit):
            _run_with_args()

    def test_custom_parsers_handle_required_attrs(self):
        class C(Corgy):
            x: Required[int]

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                return int(s)

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parsex)
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--x", type=str, action=_parser_action, required=True
        )

    def test_custom_parser_allows_cmdline_parsing_with_self_type(self):
        class C(Corgy):
            x: int
            c: Self

            @corgyparser("c")
            @staticmethod
            def parsec(s):
                return C(x=int(s))

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "1", "--c", "2"])

        c = C.parse_from_cmdline(parser)
        self.assertEqual(c, C(x=1, c=C(x=2)))

    def test_custom_parser_allows_cmdline_parsing_with_nested_self_type(self):
        class C(Corgy):
            x: int
            tc: Tuple[Self]

            @corgyparser("tc")
            @staticmethod
            def parsetc(s):
                return tuple(C(x=int(si)) for si in s.split(":"))

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(
            parser, ["--x", "1", "--tc", "2:3:4"]
        )

        c = C.parse_from_cmdline(parser)
        self.assertEqual(c, C(x=1, tc=(C(x=2), C(x=3), C(x=4))))

    def test_custom_parser_allows_cmdline_parsing_heterogenous_collection(self):
        class C(Corgy):
            x: Tuple[int, float, str]

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                s = s.split(":")
                return (int(s[0]), float(s[1]), s[2])

        parser = ArgumentParser()
        orig_parse_args = ArgumentParser.parse_args
        parser.parse_args = lambda: orig_parse_args(parser, ["--x", "2:3.0:4"])

        c = C.parse_from_cmdline(parser)
        self.assertEqual(c, C(x=(2, 3.0, "4")))

    def test_custom_parser_allows_cmdline_parsing_heterogenous_literal(self):
        class C(Corgy):
            x: Literal[1, "2"]

            @corgyparser("x")
            @staticmethod
            def parsex(s):
                if s == "1":
                    return 1
                if s == "2":
                    return "2"
                raise ValueError(s)

        for val in [1, "2"]:
            with self.subTest(val=val):
                valstr = str(val)
                parser = ArgumentParser()
                orig_parse_args = ArgumentParser.parse_args
                # pylint: disable=cell-var-from-loop
                parser.parse_args = lambda: orig_parse_args(parser, ["--x", valstr])

                c = C.parse_from_cmdline(parser)
                self.assertEqual(c, C(x=val))

    def test_add_args_uses_correct_metavar_with_custom_parser_and_choices(self):
        class T:
            __metavar__ = "custom_t"

        class T1(T):
            ...

        class T2(T):
            ...

        class C(Corgy):
            t: Literal[T1, T2]  # type: ignore

            @corgyparser("t")
            @staticmethod
            def parset(s):
                if s == "t1":
                    return T1
                if s == "t2":
                    return T2
                raise ValueError(s)

        parser = ArgumentParser()
        parser.add_argument = MagicMock()
        _parser_action = partial(CorgyParserAction, C.parset, [T1, T2])
        with patch("corgy._corgy.partial", MagicMock(return_value=_parser_action)):
            C.add_args_to_parser(parser)
        parser.add_argument.assert_called_once_with(
            "--t",
            type=str,
            action=_parser_action,
            metavar="custom_t",
            default=argparse.SUPPRESS,
        )
